from .sqs_receive import receive


class SqsService(object):
    name = "sqs-service"

    @receive('https://sqs.eu-west-1.amazonaws.com/123456789012/nameko-sqs')
    def handle_sqs_message(self, body):
        """ This method is called by the `receive` entrypoint whenever
        a message sent to the given SQS queue.
        """
        print(body)
        return body



import itertools
import time

import eventlet
import pytest
from mock import Mock, call

from nameko.events import event_handler
from nameko.exceptions import ExtensionNotFound, MethodNotFound
from nameko.extensions import DependencyProvider
from nameko.rpc import ServiceRpc, rpc
from nameko.standalone.events import event_dispatcher
from nameko.standalone.rpc import ServiceRpcClient
from nameko.testing.services import (
    entrypoint_hook, entrypoint_waiter, once, replace_dependencies,
    restrict_entrypoints, worker_factory
)
from nameko.testing.utils import get_container
from nameko.testing.waiting import wait_for_call


class LanguageReporter(DependencyProvider):
    """ Return the language given in the worker context data
    """
    def get_dependency(self, worker_ctx):
        def get_language():
            return worker_ctx.data['language']
        return get_language


handle_event = Mock()


@pytest.fixture
def counter():

    class Counter(object):
        value = 0

        def count(self):
            self.value += 1
            return self.value

    return Counter()


@pytest.yield_fixture(autouse=True)
def reset_mock():
    yield
    handle_event.reset_mock()


@pytest.yield_fixture
def spawn_thread():

    threads = []

    def spawn(fn, *args):
        """ Spawn a new thread to execute `fn(*args)`.

        The thread will be killed at test teardown if it's still running.
        """
        threads.append(eventlet.spawn(fn, *args))

    yield spawn

    for gt in threads:
        try:
            gt.kill()
        except Exception:  # pragma: no cover
            pass


class Service(object):
    name = "service"

    a = ServiceRpc("service_a")
    language = LanguageReporter()

    @rpc
    def working(self, value):
        return self.a.remote_method(value)

    @rpc
    def broken(self, value):
        raise ExampleError('broken')

    @event_handler('srcservice', 'eventtype')
    def handle(self, msg):
        handle_event(msg)

    @rpc
    def get_language(self):
        return self.language()


class ServiceA(object):

    name = "service_a"
    b = ServiceRpc("service_b")

    @rpc
    def remote_method(self, value):
        res = "{}-a".format(value)
        return self.b.remote_method(res)


class ServiceB(object):

    name = "service_b"
    c = ServiceRpc("service_c")

    @rpc
    def remote_method(self, value):
        res = "{}-b".format(value)
        return self.c.remote_method(res)


class ServiceC(object):

    name = "service_c"

    @rpc
    def remote_method(self, value):
        return "{}-c".format(value)


class ExampleError(Exception):
    pass


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_hook(runner_factory):

    service_classes = (Service, ServiceA, ServiceB, ServiceC)
    runner = runner_factory(*service_classes)
    runner.start()

    service_container = get_container(runner, Service)

    event_payload = "msg"
    with entrypoint_hook(service_container, 'handle') as handle:
        with entrypoint_waiter(service_container, 'handle'):
            handle(event_payload)
    handle_event.assert_called_once_with(event_payload)


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_hook_with_return(runner_factory):

    service_classes = (Service, ServiceA, ServiceB, ServiceC)
    runner = runner_factory(*service_classes)
    runner.start()

    service_container = get_container(runner, Service)

    with entrypoint_hook(service_container, 'working') as working:
        assert working("value") == "value-a-b-c"

    with entrypoint_hook(service_container, 'broken') as broken:
        with pytest.raises(ExampleError):
            broken("value")


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.parametrize("context_data",
                         [{'language': 'en'}, {'language': 'fr'}])
def test_entrypoint_hook_context_data(container_factory, context_data):

    container = container_factory(Service)
    container.start()

    method = 'get_language'
    with entrypoint_hook(container, method, context_data) as get_language:
        assert get_language() == context_data['language']


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_hook_dependency_not_found(container_factory):

    container = container_factory(Service)
    container.start()

    method = 'nonexistent_method'

    with pytest.raises(ExtensionNotFound):
        with entrypoint_hook(container, method):
            pass    # pragma: no cover


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_hook_container_dying(container_factory):
    class DependencyError(Exception):
        pass

    class BadDependency(DependencyProvider):
        def worker_setup(self, worker_ctx):
            raise DependencyError("Boom")

    class BadService(Service):
        bad = BadDependency()

    container = container_factory(BadService)
    container.start()

    with pytest.raises(DependencyError):
        with entrypoint_hook(container, 'working') as call:
            call()


def test_entrypoint_hook_timeout(container_factory, rabbit_config):

    class Service:
        name = 'service'

        @rpc
        def long_task(self):
            time.sleep(0.1)

    container = container_factory(Service, rabbit_config)
    container.start()

    with pytest.raises(entrypoint_waiter.Timeout) as exc_info:
        with entrypoint_hook(container, 'long_task', timeout=0.01) as call:
            call()

    assert str(exc_info.value) == (
        'Timeout on service.long_task after 0.01 seconds'
    )


def test_worker_factory():

    class Service(object):
        name = "service"
        foo_client = ServiceRpc("foo_service")
        bar_client = ServiceRpc("bar_service")

    class OtherService(object):
        pass

    # simplest case, no overrides
    instance = worker_factory(Service)
    assert isinstance(instance, Service)
    assert isinstance(instance.foo_client, Mock)
    assert isinstance(instance.bar_client, Mock)

    # no dependencies to replace
    instance = worker_factory(OtherService)
    assert isinstance(instance, OtherService)

    # override specific dependency
    bar_dependency = object()
    instance = worker_factory(Service, bar_client=bar_dependency)
    assert isinstance(instance, Service)
    assert isinstance(instance.foo_client, Mock)
    assert instance.bar_client is bar_dependency

    # non-applicable dependency
    with pytest.raises(ExtensionNotFound):
        worker_factory(Service, nonexist=object())


@pytest.mark.usefixtures("rabbit_config")
def test_replace_dependencies_kwargs(container_factory):

    class Service(object):
        name = "service"
        foo_client = ServiceRpc("foo_service")
        bar_client = ServiceRpc("bar_service")
        baz_client = ServiceRpc("baz_service")

        @rpc
        def method(self, arg):
            self.foo_client.remote_method(arg)

    class FakeDependency(object):
        def __init__(self):
            self.processed = []

        def remote_method(self, arg):
            self.processed.append(arg)

    container = container_factory(Service)

    # customise a single dependency
    fake_foo_client = FakeDependency()
    replace_dependencies(container, foo_client=fake_foo_client)
    assert 2 == len([dependency for dependency in container.extensions
                     if isinstance(dependency, ServiceRpc)])

    # customise multiple dependencies
    res = replace_dependencies(container, bar_client=Mock(), baz_client=Mock())
    assert list(res) == []

    # verify that container.extensions doesn't include an ServiceRpc anymore
    assert all([not isinstance(dependency, ServiceRpc)
                for dependency in container.extensions])

    container.start()

    # verify that the fake dependency collected calls
    msg = "msg"
    with ServiceRpcClient("service") as service_client:
        service_client.method(msg)

    assert fake_foo_client.processed == [msg]


@pytest.mark.usefixtures("rabbit_config")
def test_replace_dependencies_args(container_factory):

    class Service(object):
        name = "service"
        foo_client = ServiceRpc("foo_service")
        bar_client = ServiceRpc("bar_service")
        baz_client = ServiceRpc("baz_service")

        @rpc
        def method(self, arg):
            self.foo_client.remote_method(arg)

    container = container_factory(Service)

    # replace a single dependency
    foo_client = replace_dependencies(container, "foo_client")

    # replace multiple dependencies
    replacements = replace_dependencies(container, "bar_client", "baz_client")
    assert len([x for x in replacements]) == 2

    # verify that container.extensions doesn't include an ServiceRpc anymore
    assert all([not isinstance(dependency, ServiceRpc)
                for dependency in container.extensions])

    container.start()

    # verify that the mock dependency collects calls
    msg = "msg"
    with ServiceRpcClient("service") as service_client:
        service_client.method(msg)

    foo_client.remote_method.assert_called_once_with(msg)


@pytest.mark.usefixtures("rabbit_config")
def test_replace_dependencies_args_and_kwargs(container_factory):
    class Service(object):
        name = "service"
        foo_client = ServiceRpc("foo_service")
        bar_client = ServiceRpc("bar_service")
        baz_client = ServiceRpc("baz_service")

        @rpc
        def method(self, arg):
            self.foo_client.remote_method(arg)
            self.bar_client.bar()
            self.baz_client.baz()

    class FakeDependency(object):
        def __init__(self):
            self.processed = []

        def remote_method(self, arg):
            self.processed.append(arg)

    container = container_factory(Service)

    fake_foo_client = FakeDependency()
    mock_bar_client, mock_baz_client = replace_dependencies(
        container, 'bar_client', 'baz_client', foo_client=fake_foo_client
    )

    # verify that container.extensions doesn't include an ServiceRpc anymore
    assert all([not isinstance(dependency, ServiceRpc)
                for dependency in container.extensions])

    container.start()

    # verify that the fake dependency collected calls
    msg = "msg"
    with ServiceRpcClient("service") as service_client:
        service_client.method(msg)

    assert fake_foo_client.processed == [msg]
    assert mock_bar_client.bar.call_count == 1
    assert mock_baz_client.baz.call_count == 1


@pytest.mark.usefixtures("rabbit_config")
def test_replace_dependencies_in_both_args_and_kwargs_error(
    container_factory
):
    class Service(object):
        name = "service"
        foo_client = ServiceRpc("foo_service")
        bar_client = ServiceRpc("bar_service")
        baz_client = ServiceRpc("baz_service")

    container = container_factory(Service)

    with pytest.raises(RuntimeError) as exc:
        replace_dependencies(
            container, 'bar_client', 'foo_client', foo_client='foo'
        )
    assert "Cannot replace the same dependency" in str(exc)


@pytest.mark.usefixtures("rabbit_config")
def test_replace_non_dependency(container_factory):

    class Service(object):
        name = "service"
        client = ServiceRpc("foo_service")

        @rpc
        def method(self):
            pass  # pragma: no cover

    container = container_factory(Service)

    # error if dependency doesn't exit
    with pytest.raises(ExtensionNotFound):
        replace_dependencies(container, "nonexist")

    # error if dependency is not an dependency
    with pytest.raises(ExtensionNotFound):
        replace_dependencies(container, "method")


@pytest.mark.usefixtures("rabbit_config")
def test_replace_dependencies_container_already_started(container_factory):

    class Service(object):
        name = "service"
        client = ServiceRpc("foo_service")

    container = container_factory(Service)
    container.start()

    with pytest.raises(RuntimeError):
        replace_dependencies(container, "client")


@pytest.mark.usefixtures("rabbit_config")
def test_restrict_entrypoints(container_factory):

    method_called = Mock()

    class Service(object):
        name = "service"

        @rpc
        @once("assert not seen")
        def handler_one(self, arg):
            method_called(arg)  # pragma: no cover

        @event_handler('srcservice', 'eventtype')
        def handler_two(self, msg):
            method_called(msg)

    container = container_factory(Service)

    # disable the entrypoints on handler_one
    restrict_entrypoints(container, "handler_two")
    container.start()

    # verify the rpc entrypoint on handler_one is disabled
    with ServiceRpcClient("service") as service_client:
        with pytest.raises(MethodNotFound) as exc_info:
            service_client.handler_one("msg")
        assert str(exc_info.value) == "handler_one"

    # dispatch an event to handler_two
    msg = "msg"
    dispatch = event_dispatcher()

    with entrypoint_waiter(container, 'handler_two'):
        dispatch('srcservice', 'eventtype', msg)

    # method_called should have exactly one call, derived from the event
    # handler and not from the disabled @once entrypoint
    method_called.assert_called_once_with(msg)


@pytest.mark.usefixtures("rabbit_config")
def test_restrict_nonexistent_entrypoint(container_factory):

    class Service(object):
        name = "service"

        @rpc
        def method(self, arg):
            pass  # pragma: no cover

    container = container_factory(Service)

    with pytest.raises(ExtensionNotFound):
        restrict_entrypoints(container, "nonexist")


@pytest.mark.usefixtures("rabbit_config")
def test_restrict_entrypoint_container_already_started(container_factory):

    class Service(object):
        name = "service"

        @rpc
        def method(self, arg):
            pass  # pragma: no cover

    container = container_factory(Service)
    container.start()

    with pytest.raises(RuntimeError):
        restrict_entrypoints(container, "method")


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter(container_factory):
    container = container_factory(Service)
    container.start()

    dispatch = event_dispatcher()
    with entrypoint_waiter(container, 'handle'):
        dispatch('srcservice', 'eventtype', "")


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter_result(container_factory):

    class Service(object):
        name = "service"

        @event_handler('srcservice', 'eventtype')
        def handle_event(self, msg):
            return msg.upper()

    container = container_factory(Service)
    container.start()

    dispatch = event_dispatcher()
    with entrypoint_waiter(container, 'handle_event') as result:
        dispatch('srcservice', 'eventtype', "msg")

    res = result.get()
    assert res == "MSG"


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter_with_callback(container_factory):

    class Service(object):
        name = "service"

        @event_handler('srcservice', 'eventtype')
        def handle_event(self, msg):
            return msg

    container = container_factory(Service)
    container.start()

    results = []

    def cb(worker_ctx, res, exc_info):
        results.append((res, exc_info))
        return len(results) == 2

    dispatch = event_dispatcher()
    with entrypoint_waiter(container, 'handle_event', callback=cb):
        dispatch('srcservice', 'eventtype', "msg1")
        dispatch('srcservice', 'eventtype', "msg2")

    assert results == [("msg1", None), ("msg2", None)]


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter_wait_for_specific_result(
    container_factory, spawn_thread
):

    class Service(object):
        name = "service"

        @event_handler('srcservice', 'eventtype')
        def handle_event(self, msg):
            return msg

    container = container_factory(Service)
    container.start()

    target = 5

    def cb(worker_ctx, res, exc_info):
        return res == target

    def increment_forever():
        dispatch = event_dispatcher()
        for count in itertools.count():
            dispatch('srcservice', 'eventtype', count)
            time.sleep()  # force yield

    with entrypoint_waiter(container, 'handle_event', callback=cb) as result:
        spawn_thread(increment_forever)

    assert result.get() == target


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter_wait_until_called_with_argument(
    container_factory, spawn_thread
):

    class Service(object):
        name = "service"

        @event_handler('srcservice', 'eventtype')
        def handle_event(self, msg):
            return msg

    container = container_factory(Service)
    container.start()

    target = 5

    def cb(worker_ctx, res, exc_info):
        return worker_ctx.args == (target,)

    def increment_forever():
        dispatch = event_dispatcher()
        for count in itertools.count():
            dispatch('srcservice', 'eventtype', count)
            time.sleep()  # force yield

    with entrypoint_waiter(container, 'handle_event', callback=cb) as result:
        spawn_thread(increment_forever)

    assert result.get() == target


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter_wait_until_raises(
    container_factory, spawn_thread
):
    threshold = 5

    class TooMuch(Exception):
        pass

    class Service(object):
        name = "service"

        @event_handler('srcservice', 'eventtype')
        def handle_event(self, msg):
            if msg > threshold:
                raise TooMuch(msg)
            return msg

    container = container_factory(Service)
    container.start()

    def cb(worker_ctx, res, exc_info):
        return exc_info is not None

    def increment_forever():
        dispatch = event_dispatcher()
        for count in itertools.count():
            dispatch('srcservice', 'eventtype', count)
            time.sleep()  # force yield

    with entrypoint_waiter(container, 'handle_event', callback=cb) as result:
        spawn_thread(increment_forever)

    with pytest.raises(TooMuch):
        result.get()


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter_wait_until_stops_raising(
    container_factory, spawn_thread
):
    threshold = 5

    class NotEnough(Exception):
        pass

    class Service(object):
        name = "service"

        @event_handler('srcservice', 'eventtype')
        def handle_event(self, msg):
            if msg < threshold:
                raise NotEnough(msg)
            return msg

    container = container_factory(Service)
    container.start()

    def cb(worker_ctx, res, exc_info):
        return exc_info is None

    def increment_forever():
        dispatch = event_dispatcher()
        for count in itertools.count():
            dispatch('srcservice', 'eventtype', count)
            time.sleep()  # force yield

    with entrypoint_waiter(container, 'handle_event', callback=cb) as result:
        spawn_thread(increment_forever)

    assert result.get() == threshold


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter_timeout(container_factory):
    container = container_factory(Service)
    container.start()

    with pytest.raises(entrypoint_waiter.Timeout) as exc_info:
        with entrypoint_waiter(container, 'handle', timeout=0.01):
            pass
    assert str(exc_info.value) == (
        "Timeout on service.handle after 0.01 seconds")


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter_bad_entrypoint(container_factory):
    container = container_factory(Service)

    with pytest.raises(RuntimeError) as exc:
        with entrypoint_waiter(container, "unknown"):
            pass  # pragma: no cover
    assert 'has no entrypoint' in str(exc)


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter_nested(container_factory):

    class Service(object):
        name = "service"

        @event_handler('srcservice', 'eventtype1')
        def handle_event1(self, msg):
            handle_event(1)

        @event_handler('srcservice', 'eventtype2')
        def handle_event2(self, msg):
            handle_event(2)

    container = container_factory(Service)
    container.start()

    dispatch = event_dispatcher()
    with entrypoint_waiter(container, 'handle_event1'):
        with entrypoint_waiter(container, 'handle_event2'):
            dispatch('srcservice', 'eventtype1', "")
            dispatch('srcservice', 'eventtype2', "")

    assert call(1) in handle_event.call_args_list
    assert call(2) in handle_event.call_args_list


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter_duplicate(container_factory):

    class Service(object):
        name = "service"

        @event_handler('srcservice', 'eventtype')
        def handle_event(self, msg):
            handle_event(msg)

    container = container_factory(Service)
    container.start()

    dispatch = event_dispatcher()
    with entrypoint_waiter(container, 'handle_event'):
        with entrypoint_waiter(container, 'handle_event'):
            dispatch('srcservice', 'eventtype', "msg")

    assert handle_event.call_args_list == [call("msg")]


@pytest.mark.usefixtures("rabbit_config")
def test_entrypoint_waiter_result_teardown_race(
    container_factory, counter
):
    tracker = Mock()

    class TrackingDependency(DependencyProvider):

        def worker_result(self, worker_ctx, res, exc_info):
            tracker.worker_result()

        def worker_teardown(self, worker_ctx):
            tracker.worker_teardown()

    class Service(object):
        name = "service"

        tracker = TrackingDependency()

        @event_handler('srcservice', 'eventtype')
        def handle(self, msg):
            tracker.handle(msg)

    container = container_factory(Service)
    container.start()

    def wait_for_two_calls(worker_ctx, res, exc_info):
        return counter.count() > 1

    dispatch = event_dispatcher()
    with entrypoint_waiter(container, 'handle', callback=wait_for_two_calls):

        # dispatch the first message
        # wait until teardown has happened
        with wait_for_call(TrackingDependency, 'worker_teardown'):
            dispatch('srcservice', 'eventtype', "msg")

        assert tracker.worker_teardown.call_count == 1
        assert tracker.worker_result.call_count == 1
        assert tracker.handle.call_count == 1

        # dispatch the second event
        dispatch('srcservice', 'eventtype', "msg")

    # we should wait for the second teardown to complete before exiting
    # the entrypoint waiter
    assert tracker.worker_teardown.call_count == 2
    assert tracker.worker_result.call_count == 2
    assert tracker.handle.call_count == 2



'''
Provides core messaging decorators and dependency providers.
'''
from __future__ import absolute_import

import re
from functools import partial
from logging import getLogger

from kombu.common import maybe_declare

from nameko import config, serialization
from nameko.amqp.consume import Consumer as ConsumerCore
from nameko.amqp.publish import Publisher as PublisherCore
from nameko.amqp.publish import get_connection
from nameko.constants import (
    AMQP_SSL_CONFIG_KEY, AMQP_URI_CONFIG_KEY, DEFAULT_AMQP_URI,
    DEFAULT_HEARTBEAT, DEFAULT_PREFETCH_COUNT, HEADER_PREFIX,
    HEARTBEAT_CONFIG_KEY, PREFETCH_COUNT_CONFIG_KEY
)
from nameko.exceptions import ContainerBeingKilled
from nameko.extensions import DependencyProvider, Entrypoint


_log = getLogger(__name__)


def encode_to_headers(context_data, prefix=HEADER_PREFIX):
    return {
        "{}.{}".format(prefix, key): value
        for key, value in context_data.items()
        if value is not None
    }


def decode_from_headers(headers, prefix=HEADER_PREFIX):
    return {
        re.sub("^{}\.".format(prefix), "", key): value
        for key, value in headers.items()
    }


class Publisher(DependencyProvider):

    publisher_cls = PublisherCore

    def __init__(self, exchange=None, declare=None, **publisher_options):
        """ Provides an AMQP message publisher method via dependency injection.

        In AMQP, messages are published to *exchanges* and routed to bound
        *queues*. This dependency accepts the `exchange` to publish to and
        will ensure that it is declared before publishing.

        Optionally, you may use the `declare` keyword argument to pass a list
        of other :class:`kombu.Exchange` or :class:`kombu.Queue` objects to
        declare before publishing.

        :Parameters:
            exchange : :class:`kombu.Exchange`
                Destination exchange
            declare : list
                List of :class:`kombu.Exchange` or :class:`kombu.Queue` objects
                to declare before publishing.
            **publisher_options
                Options to configure the
                :class:`~nameko.amqqp.publish.Publisher` that sends the
                message.


        If `exchange` is not provided, the message will be published to the
        default exchange.

        Example::

            class Foobar(object):

                publish = Publisher(exchange=...)

                def spam(self, data):
                    self.publish('spam:' + data)
        """
        self.exchange = exchange
        self.publisher_options = publisher_options

        self.declare = declare[:] if declare is not None else []

        if self.exchange:
            self.declare.append(self.exchange)

        default_uri = config.get(AMQP_URI_CONFIG_KEY, DEFAULT_AMQP_URI)
        self.amqp_uri = self.publisher_options.pop('uri', default_uri)

    def setup(self):

        default_ssl = config.get(AMQP_SSL_CONFIG_KEY)
        ssl = self.publisher_options.pop('ssl', default_ssl)

        with get_connection(self.amqp_uri, ssl) as conn:
            for entity in self.declare:
                maybe_declare(entity, conn.channel())

        default_serializer = self.container.serializer
        serializer = self.publisher_options.pop(
            'serializer', default_serializer
        )

        self.publisher = self.publisher_cls(
            self.amqp_uri,
            ssl=ssl,
            serializer=serializer,
            exchange=self.exchange,
            declare=self.declare,
            **self.publisher_options
        )

    def get_dependency(self, worker_ctx):

        def publish(msg, **kwargs):
            extra_headers = encode_to_headers(worker_ctx.context_data)
            self.publisher.publish(
                msg, extra_headers=extra_headers, **kwargs
            )

        return publish


class Consumer(Entrypoint):

    consumer_cls = ConsumerCore

    def __init__(
        self, queue, requeue_on_error=False,
        expected_exceptions=(), sensitive_arguments=(), **consumer_options
    ):
        """
        Decorates a method as a message consumer.

        Messages from the queue will be deserialized depending on their content
        type and passed to the the decorated method.
        When the consumer method returns without raising any exceptions,
        the message will automatically be acknowledged.
        If any exceptions are raised during the consumption and
        `requeue_on_error` is True, the message will be requeued.

        If `requeue_on_error` is true, handlers will return the event to the
        queue if an error occurs while handling it. Defaults to false.

        Example::

            @consume(...)
            def handle_message(self, body):

                if not self.spam(body):
                    raise Exception('message will be requeued')

                self.shrub(body)

        Args:
            queue: The queue to consume from.
        """
        self.queue = queue
        self.requeue_on_error = requeue_on_error
        self.consumer_options = consumer_options
        # TODO it's bad that we eat all the remaining keyword arguments as consumer opts
        super(Consumer, self).__init__(
            expected_exceptions=expected_exceptions,
            sensitive_arguments=sensitive_arguments
        )

    @property
    def amqp_uri(self):
        return config[AMQP_URI_CONFIG_KEY]

    def setup(self):
        ssl = config.get(AMQP_SSL_CONFIG_KEY)

        heartbeat = self.consumer_options.pop(
            'heartbeat', config.get(HEARTBEAT_CONFIG_KEY, DEFAULT_HEARTBEAT)
        )
        prefetch_count = self.consumer_options.pop(
            'prefetch_count', config.get(
                PREFETCH_COUNT_CONFIG_KEY, DEFAULT_PREFETCH_COUNT
            )
        )
        accept = self.consumer_options.pop(
            'accept', serialization.setup().accept
        )

        queues = [self.queue]
        callbacks = [self.handle_message]

        self.consumer = self.consumer_cls(
            self.amqp_uri, ssl=ssl, queues=queues, callbacks=callbacks,
            heartbeat=heartbeat, prefetch_count=prefetch_count, accept=accept,
            **self.consumer_options
        )

    def start(self):
        self.container.spawn_managed_thread(self.consumer.run)
        self.consumer.wait_until_consumer_ready()

    def stop(self):
        self.consumer.stop()

    def handle_message(self, body, message):
        args = (body,)
        kwargs = {}

        context_data = decode_from_headers(message.headers)

        handle_result = partial(self.handle_result, message)

        def spawn_worker():
            try:
                self.container.spawn_worker(
                    self, args, kwargs,
                    context_data=context_data,
                    handle_result=handle_result
                )
            except ContainerBeingKilled:
                self.consumer.requeue_message(message)

        service_name = self.container.service_name
        method_name = self.method_name

        # we must spawn a thread here to prevent handle_message blocking
        # when the worker pool is exhausted; if this happens the AMQP consumer
        # is also blocked and fails to send heartbeats, eventually causing it
        # to be disconnected
        # TODO replace global worker pool limits with per-entrypoint limits,
        # then remove this waiter thread
        ident = u"{}.wait_for_worker_pool[{}.{}]".format(
            type(self).__name__, service_name, method_name
        )
        self.container.spawn_managed_thread(spawn_worker, identifier=ident)

    def handle_result(self, message, worker_ctx, result=None, exc_info=None):
        self.handle_message_processed(message, result, exc_info)
        return result, exc_info

    def handle_message_processed(self, message, result=None, exc_info=None):

        if exc_info is not None and self.requeue_on_error:
            self.consumer.requeue_message(message)
        else:
            self.consumer.ack_message(message)


consume = Consumer.decorator



from logging import getLogger
from threading import Event

from amqp.exceptions import ConnectionError
from kombu import Connection
from kombu.mixins import ConsumerMixin

from nameko.utils import sanitize_url


_log = getLogger(__name__)


class Consumer(ConsumerMixin):
    """ Helper utility for consuming messages from RabbitMQ.
    """

    def __init__(
        self, amqp_uri, ssl=None, queues=None, callbacks=None, heartbeat=None,
        prefetch_count=None, accept=None, **consumer_options
    ):
        self.amqp_uri = amqp_uri
        self.ssl = ssl

        self.queues = queues
        self.callbacks = callbacks or []
        self.heartbeat = heartbeat
        self.prefetch_count = prefetch_count or 0
        self.accept = accept

        self.consumer_options = consumer_options

        self.ready = Event()

        super(Consumer, self).__init__()

    @property
    def connection(self):
        """ Provide the connection parameters for kombu's ConsumerMixin.

        The `Connection` object is a declaration of connection parameters
        that is lazily evaluated. It doesn't represent an established
        connection to the broker at this point.
        """
        return Connection(
            self.amqp_uri, ssl=self.ssl, heartbeat=self.heartbeat
        )

    def stop(self):
        """ Stop this consumer.

        Any messages received between when this method is called and the
        resulting consumer cancel will be requeued.
        """
        self.should_stop = True

    def wait_until_consumer_ready(self):
        """ Wait for initial connection.
        """
        self.ready.wait()

    def get_consumers(self, consumer_cls, channel):
        """ Kombu callback to set up consumers.

        Called after any (re)connection to the broker.
        """
        consumer = consumer_cls(
            queues=self.queues,
            callbacks=[self.on_message],
            accept=self.accept,
            **self.consumer_options
        )
        consumer.qos(prefetch_count=self.prefetch_count)
        return [consumer]

    def on_consume_ready(self, connection, channel, consumers, **kwargs):
        """ Kombu callback when consumers are ready to accept messages.

        Called after any (re)connection to the broker.
        """
        # if this is the very first connection, send the event so anything
        # blocking on this Consumer "starting" can continue
        if not self.ready.is_set():
            self.ready.set()

    def on_connection_error(self, exc, interval):
        _log.warning(
            "Error connecting to broker at {} ({}).\n"
            "Retrying in {} seconds."
            .format(sanitize_url(self.amqp_uri), exc, interval)
        )

    def on_message(self, body, message):
        if self.should_stop:
            self.requeue_message(message)
            return
        for callback in self.callbacks:
            callback(body, message)

    def ack_message(self, message):
        # only attempt to ack if the message connection is alive;
        # otherwise the message will already have been reclaimed by the broker
        if message.channel.connection:
            try:
                message.ack()
            except ConnectionError:  # pragma: no cover
                pass  # ignore connection closing inside conditional

    def requeue_message(self, message):
        # only attempt to requeue if the message connection is alive;
        # otherwise the message will already have been reclaimed by the broker
        if message.channel.connection:
            try:
                message.requeue()
            except ConnectionError:  # pragma: no cover
                pass  # ignore connection closing inside conditional



import errno
import json
import socket

import eventlet
import pytest
from eventlet.event import Event

from nameko.exceptions import (
    MalformedRequest, MethodNotFound, RemoteError, deserialize
)
from nameko.testing.services import dummy, entrypoint_hook, get_extension
from nameko.testing.websocket import make_virtual_socket
from nameko.web.websocket import WebSocketHubProvider, WebSocketRpc, rpc


class ExampleService(object):
    name = "exampleservice"

    websocket_hub = WebSocketHubProvider()

    @rpc
    def subscribe(self, socket_id):
        self.websocket_hub.subscribe(socket_id, 'test_channel')
        return 'subscribed!'

    @rpc
    def unsubscribe(self, socket_id):
        self.websocket_hub.unsubscribe(socket_id, 'test_channel')
        return 'unsubscribed!'

    @rpc
    def list_subscriptions(self, socket_id):
        return self.websocket_hub.get_subscriptions(socket_id)

    @dummy
    def broadcast(self, value):
        self.websocket_hub.broadcast('test_channel', 'test_message', {
            'value': value,
        })

    @dummy
    def unicast(self, target_socket_id, value):
        status = self.websocket_hub.unicast(target_socket_id, 'test_message', {
            'value': value,
        })
        return status

    @dummy
    def list_subscribers(self):
        return self.websocket_hub.subscriptions


def get_message(ws):
    # matches the broadcast rpc call
    event_type, event_data = ws.wait_for_event('test_message')
    assert event_type == 'test_message'
    assert list(event_data) == ['value']
    return event_data['value']


@pytest.yield_fixture
def container(container_factory, web_config):
    container = container_factory(ExampleService)
    container.start()
    yield container


def test_pub_sub(container, websocket):
    ws = websocket()
    assert ws.rpc('subscribe') == 'subscribed!'
    with entrypoint_hook(container, 'broadcast') as broadcast:
        broadcast(value=42)

    assert get_message(ws) == 42


def test_resubscribe(container, websocket):
    ws = websocket()
    ws.rpc('subscribe')
    ws.rpc('subscribe')

    with entrypoint_hook(container, 'broadcast') as broadcast:
        broadcast(value=42)
    assert get_message(ws) == 42


def test_multiple_calls(container, websocket):
    ws = websocket()
    ws.rpc('subscribe')
    count = 2
    for value in range(count):
        with entrypoint_hook(container, 'broadcast') as broadcast:
            broadcast(value=value)

    for value in range(count):
        assert get_message(ws) == value


def test_unsubscribe(container, websocket):
    ws = websocket()
    ws.rpc('subscribe')
    with entrypoint_hook(container, 'broadcast') as broadcast:
        broadcast(value=42)
    assert get_message(ws) == 42
    ws.rpc('unsubscribe')

    with entrypoint_hook(container, 'broadcast') as broadcast:
        broadcast(value=42)
    with eventlet.Timeout(.1, exception=False):
        assert get_message(ws) == 42


def test_unsubscribe_noop(container, websocket):
    ws = websocket()
    ws.rpc('unsubscribe')

    with entrypoint_hook(container, 'broadcast') as broadcast:
        broadcast(value=42)
    with eventlet.Timeout(.1, exception=False):
        assert get_message(ws) == 42


def test_multiple_subscribers(container, websocket):
    ws1 = websocket()
    ws2 = websocket()

    ws1.rpc('subscribe')
    ws2.rpc('subscribe')

    with entrypoint_hook(container, 'broadcast') as broadcast:
        broadcast(value=42)
    assert get_message(ws1) == 42
    assert get_message(ws2) == 42


def test_close(container, websocket):
    ws = websocket()

    ws.rpc('subscribe')

    with entrypoint_hook(container, 'list_subscribers') as list_subscribers:
        subscribers1 = list_subscribers()
        assert subscribers1['test_channel']

    ws.app.close()

    with entrypoint_hook(container, 'list_subscribers') as list_subscribers:
        subscribers2 = list_subscribers()
        assert not subscribers2['test_channel']


def test_method_not_found(container, websocket):
    ws = websocket()
    with pytest.raises(MethodNotFound):
        ws.rpc('unknown')


def test_list_subscriptions(container, websocket):
    ws = websocket()
    assert ws.rpc('list_subscriptions') == []
    ws.rpc('subscribe')
    assert ws.rpc('list_subscriptions') == ['test_channel']


def test_unicast(container, websocket):
    ws = websocket()
    _, connected_data = ws.wait_for_event('connected')
    socket_id = connected_data['socket_id']
    with entrypoint_hook(container, 'unicast') as unicast:
        assert unicast(target_socket_id=socket_id, value=42)
    assert get_message(ws) == 42


def test_unicast_unknown(container):
    with entrypoint_hook(container, 'unicast') as unicast:
        assert not unicast(target_socket_id=0, value=42)


def test_connection_not_found(container, websocket):
    hub = get_extension(container, WebSocketHubProvider)
    ws = websocket()
    hub.server.sockets.clear()

    # doesn't need to be known
    assert ws.rpc('unsubscribe') == 'unsubscribed!'

    with pytest.raises(RemoteError) as exc:
        ws.rpc('subscribe')
    assert exc.value.exc_type == 'ConnectionNotFound'


def test_badly_encoded_data(container, web_config_port):
    ws_app, wait_for_sock = make_virtual_socket('127.0.0.1', web_config_port)

    gt = eventlet.spawn(ws_app.run_forever)
    wait_for_sock()
    result = Event()

    def on_message(ws, message):
        response = json.loads(message)
        assert not response['success']
        exc = deserialize(response['error'])
        result.send_exception(exc)
    ws_app.on_message = on_message
    ws_app.send('foo: bar')

    with pytest.raises(MalformedRequest) as exc:
        result.wait()
    assert 'Invalid JSON data' in str(exc)

    ws_app.close()
    gt.kill()


def test_websocket_helper_error(websocket):
    with pytest.raises(socket.error) as exc:
        websocket()
    assert exc.value.errno == errno.ECONNREFUSED


def test_client_closing_connection(container, web_config_port):
    ws_app, wait_for_sock = make_virtual_socket('127.0.0.1', web_config_port)

    gt = eventlet.spawn(ws_app.run_forever)
    wait_for_sock()

    wait_for_close = Event()

    def on_close(ws):
        wait_for_close.send(None)

    ws_app.on_close = on_close

    ws_app.send(b'\xff\x00')  # "Close the connection" packet.
    wait_for_close.wait()

    ws_app.close()
    gt.kill()

    assert container.stop() is None


class TestEntrypointArguments:

    @pytest.mark.usefixtures("web_config")
    def test_expected_exceptions_and_sensitive_arguments(
        self, container_factory
    ):

        class Boom(Exception):
            pass

        class Service(object):
            name = "service"

            @rpc(expected_exceptions=Boom, sensitive_arguments=["arg"])
            def method(self, arg):
                pass  # pragma: no cover

        container = container_factory(Service)
        container.start()

        entrypoint = get_extension(container, WebSocketRpc)
        assert entrypoint.expected_exceptions == Boom
        assert entrypoint.sensitive_arguments == ["arg"]



from nameko.containers import ServiceContainer

class Service:
    name = "service"

# create a container
container = ServiceContainer(Service)

# ``container.extensions`` exposes all extensions used by the service
service_extensions = list(container.extensions)

# start service
container.start()

# stop service
container.stop()



from nameko.rpc import rpc

from .auth import Auth, Unauthorized


class Service:
    name = "service"

    auth = Auth()

    @rpc(expected_exceptions=Unauthorized)
    def update(self, data):
        if not self.auth.has_role("admin"):
            raise Unauthorized()

        # perform update
        raise TypeError("Whoops, genuine error.")



from collections import defaultdict

import pytest
from mock import Mock, call

from nameko.extensions import DependencyProvider, Entrypoint
from nameko.testing.services import dummy, entrypoint_hook, once
from nameko.testing.utils import get_extension
from nameko.utils import REDACTED, get_redacted_args


@pytest.fixture
def tracker():
    return Mock()


class TestDecorator(object):

    def test_decorator_without_args(self, container_factory, tracker):

        class Service(object):
            name = "service"

            @once
            def method(self, a="a", b="b"):
                tracker(a, b)

        container = container_factory(Service)
        container.start()
        container.stop()  # graceful, waits for workers

        assert tracker.call_args == call('a', 'b')

    def test_decorator_with_args(self, container_factory, tracker):

        class Service(object):
            name = "service"

            @once("x")
            def method(self, a, b="b"):
                tracker(a, b)

        container = container_factory(Service)
        container.start()
        container.stop()  # graceful, waits for workers

        assert tracker.call_args == call('x', 'b')

    def test_decorator_with_kwargs(self, container_factory, tracker):

        class Service(object):
            name = "service"

            @once(b="x")
            def method(self, a="a", b="b"):
                tracker(a, b)

        container = container_factory(Service)
        container.start()
        container.stop()  # graceful, waits for workers

        assert tracker.call_args == call('a', 'x')


class TestExpectedExceptions(object):

    def test_expected_exceptions(self, container_factory):

        exceptions = defaultdict(list)

        class CustomException(Exception):
            pass

        class Logger(DependencyProvider):
            """ Example DependencyProvider that interprets
            ``expected_exceptions`` on an entrypoint
            """

            def worker_result(self, worker_ctx, result=None, exc_info=None):
                if exc_info is None:  # nothing to do
                    return  # pragma: no cover

                exc = exc_info[1]
                expected = worker_ctx.entrypoint.expected_exceptions

                if isinstance(exc, expected):
                    exceptions['expected'].append(exc)
                else:
                    exceptions['unexpected'].append(exc)

        class Service(object):
            name = "service"

            logger = Logger()

            @dummy(expected_exceptions=CustomException)
            def expected(self):
                raise CustomException()

            @dummy
            def unexpected(self):
                raise CustomException()

        container = container_factory(Service)
        container.start()

        with entrypoint_hook(container, 'expected') as hook:
            with pytest.raises(CustomException) as expected_exc:
                hook()
        assert expected_exc.value in exceptions['expected']

        with entrypoint_hook(container, 'unexpected') as hook:
            with pytest.raises(CustomException) as unexpected_exc:
                hook()
        assert unexpected_exc.value in exceptions['unexpected']


class TestSensitiveArguments(object):

    def test_sensitive_arguments(self, container_factory):

        redacted = {}

        class Logger(DependencyProvider):
            """ Example DependencyProvider that makes use of
            ``get_redacted_args`` to redact ``sensitive_arguments``
            on entrypoints.
            """

            def worker_setup(self, worker_ctx):
                entrypoint = worker_ctx.entrypoint
                args = worker_ctx.args
                kwargs = worker_ctx.kwargs

                redacted.update(get_redacted_args(entrypoint, *args, **kwargs))

        class Service(object):
            name = "service"

            logger = Logger()

            @dummy(sensitive_arguments=("a", "b.x[0]", "b.x[2]"))
            def method(self, a, b, c):
                return [a, b, c]

        container = container_factory(Service)
        entrypoint = get_extension(container, Entrypoint)

        assert entrypoint.sensitive_arguments == ("a", "b.x[0]", "b.x[2]")

        a = "A"
        b = {'x': [1, 2, 3], 'y': [4, 5, 6]}
        c = "C"

        with entrypoint_hook(container, "method") as method:
            assert method(a, b, c) == [a, b, c]

        assert redacted == {
            'a': REDACTED,
            'b': {
                'x': [REDACTED, 2, REDACTED],
                'y': [4, 5, 6]
            },
            'c': 'C'
        }



import requests

from nameko.extensions import DependencyProvider
from nameko.rpc import rpc

URL_TEMPLATE = "https://api.travis-ci.org/repos/{}/{}"


class ApiWrapper:

    def __init__(self, session):
        self.session = session

    def repo_status(self, owner, repo):
        url = URL_TEMPLATE.format(owner, repo)
        return self.session.get(url).json()


class TravisWebservice(DependencyProvider):

    def setup(self):
        self.session = requests.Session()

    def get_dependency(self, worker_ctx):
        return ApiWrapper(self.session)


class Travis:
    name = "travis_service"

    webservice = TravisWebservice()

    @rpc
    def status_message(self, owner, repo):
        status = self.webservice.repo_status(owner, repo)
        outcome = "passing" if status['last_build_result'] else "failing"

        return "Project {repo} {outcome} since {timestamp}.".format(
            repo=status['slug'],
            outcome=outcome,
            timestamp=status['last_build_finished_at']
        )



from __future__ import print_function

import eventlet
eventlet.monkey_patch()  # noqa (code before rest of imports)

import errno
import logging
import logging.config
import signal
import sys

from eventlet import backdoor

from nameko import config
from nameko.cli.utils import import_services
from nameko.runners import ServiceRunner


logger = logging.getLogger(__name__)


def setup_backdoor(runner, backdoor_port):
    def _bad_call():
        raise RuntimeError(
            "This would kill your service, not close the backdoor. To exit, "
            "use ctrl-c."
        )

    host, port = backdoor_port
    socket = eventlet.listen((host, port))
    # work around https://github.com/celery/kombu/issues/838
    socket.settimeout(None)
    gt = eventlet.spawn(
        backdoor.backdoor_server,
        socket,
        locals={"runner": runner, "quit": _bad_call, "exit": _bad_call},
    )
    return socket, gt


def run(service_modules, backdoor_port=None):

    if "." not in sys.path:
        sys.path.append(".")

    services = []
    for path in service_modules:
        services.extend(import_services(path))

    service_runner = ServiceRunner()
    for service_cls in services:
        service_runner.add_service(service_cls)

    def shutdown(signum, frame):
        # signal handlers are run by the MAINLOOP and cannot use eventlet
        # primitives, so we have to call `stop` in a greenlet
        eventlet.spawn_n(service_runner.stop)

    signal.signal(signal.SIGTERM, shutdown)

    if backdoor_port is not None:
        setup_backdoor(service_runner, backdoor_port)

    service_runner.start()

    # if the signal handler fires while eventlet is waiting on a socket,
    # the __main__ greenlet gets an OSError(4) "Interrupted system call".
    # This is a side-effect of the eventlet hub mechanism. To protect nameko
    # from seeing the exception, we wrap the runner.wait call in a greenlet
    # spawned here, so that we can catch (and silence) the exception.
    runnlet = eventlet.spawn(service_runner.wait)

    while True:
        try:
            runnlet.wait()
        except OSError as exc:
            if exc.errno == errno.EINTR:
                # this is the OSError(4) caused by the signalhandler.
                # ignore and go back to waiting on the runner
                continue
            raise
        except KeyboardInterrupt:
            print()  # looks nicer with the ^C e.g. bash prints in the terminal
            try:
                service_runner.stop()
            except KeyboardInterrupt:
                print()  # as above
                service_runner.kill()
        else:
            # runner.wait completed
            break


def main(services, backdoor_port):
    # sys.path already manipulated at services import time
    if "LOGGING" in config:
        logging.config.dictConfig(config["LOGGING"])
    else:
        logging.basicConfig(level=logging.INFO, format="%(message)s")

    run(services, backdoor_port=backdoor_port)



from contextlib import contextmanager

import eventlet
import pytest
from eventlet.event import Event
from kombu.connection import Connection
from kombu.exceptions import OperationalError
from kombu.message import Message
from mock import Mock, patch
from six.moves import queue

from nameko import config
from nameko.amqp.consume import Consumer
from nameko.constants import HEARTBEAT_CONFIG_KEY
from nameko.containers import WorkerContext
from nameko.exceptions import (
    RemoteError, ReplyQueueExpiredWithPendingReplies, RpcTimeout
)
from nameko.extensions import DependencyProvider
from nameko.rpc import Responder, get_rpc_exchange, rpc
from nameko.standalone.rpc import (
    ClusterRpcClient, ReplyListener, ServiceRpcClient
)
from nameko.testing.waiting import wait_for_call

from test import skip_if_no_toxiproxy


class ContextReader(DependencyProvider):
    """ Access values from the worker context data.

    This is a test facilty! Write specific Dependencies to make use of
    values in ``WorkerContext.data``, don't expose it directly.
    """

    def get_dependency(self, worker_ctx):
        def get_context_value(key):
            return worker_ctx.context_data.get(key)
        return get_context_value


class FooService(object):
    name = 'foobar'

    get_context_value = ContextReader()

    @rpc
    def spam(self, ham):
        return ham

    @rpc
    def broken(self):
        raise ExampleError('broken')

    @rpc
    def get_context_data(self, name):
        return self.get_context_value(name)

    @rpc
    def sleep(self, seconds=0):
        eventlet.sleep(seconds)
        return seconds


class ExampleError(Exception):
    pass


class CustomWorkerContext(WorkerContext):
    context_keys = ("custom_header",)


@pytest.yield_fixture
def toxic_reply_listener(toxiproxy):
    class ToxicReplyListener(ReplyListener):
        def __init__(self, *args, **kwargs):
            kwargs["uri"] = toxiproxy.uri
            return super(ToxicReplyListener, self).__init__(*args, **kwargs)
    with patch('nameko.standalone.rpc.ReplyListener', ToxicReplyListener):
        yield


@pytest.mark.usefixtures('rabbit_config')
def test_client(container_factory):

    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar') as foo:
        assert foo.spam(ham='eggs') == 'eggs'
        assert foo.spam(ham='eggs') == 'eggs'  # test re-use


@pytest.mark.usefixtures('rabbit_config')
def test_client_manual_start_stop(container_factory):

    container = container_factory(FooService)
    container.start()

    foobar_client = ServiceRpcClient('foobar')
    foo = foobar_client.start()
    assert foo.spam(ham='eggs') == 'eggs'
    assert foo.spam(ham='eggs') == 'eggs'  # test re-use
    foobar_client.stop()


@pytest.mark.usefixtures('rabbit_config')
def test_client_stop_start_again(container_factory):

    container = container_factory(FooService)
    container.start()

    foobar_client = ServiceRpcClient('foobar')
    foo1 = foobar_client.start()
    assert foo1.spam(ham='eggs') == 'eggs'
    foobar_client.stop()

    foo2 = foobar_client.start()
    assert foo2.spam(ham='eggs') == 'eggs'  # test re-use
    foobar_client.stop()


@pytest.mark.usefixtures('rabbit_config')
def test_client_context_data(container_factory):

    container = container_factory(FooService)
    container.start()

    context_data = {'language': 'en'}
    with ServiceRpcClient('foobar', context_data) as foo:
        assert foo.get_context_data('language') == 'en'

    context_data = {'language': 'fr'}
    with ServiceRpcClient('foobar', context_data) as foo:
        assert foo.get_context_data('language') == 'fr'


@pytest.mark.usefixtures('rabbit_config')
@pytest.mark.usefixtures('predictable_call_ids')
def test_call_id(container_factory):

    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar') as foo:
        stack1 = foo.get_context_data('call_id_stack')
        assert stack1 == [
            'standalone_rpc_client.0.0',
            'foobar.get_context_data.1'
        ]
        stack2 = foo.get_context_data('call_id_stack')
        assert stack2 == [
            'standalone_rpc_client.0.2',
            'foobar.get_context_data.3'
        ]


@pytest.mark.usefixtures('rabbit_config')
def test_client_remote_error(container_factory):

    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient("foobar") as client:
        with pytest.raises(RemoteError) as exc_info:
            client.broken()
        assert exc_info.value.exc_type == "ExampleError"


@pytest.mark.usefixtures('rabbit_config')
@patch('nameko.standalone.rpc.RPC_REPLY_QUEUE_TTL', new=200)
def test_reply_queue_removed_on_expiry(
    rabbit_manager, get_vhost, container_factory
):
    def list_queues():
        vhost = get_vhost(config['AMQP_URI'])
        return [
            queue['name']
            for queue in rabbit_manager.get_queues(vhost=vhost)
        ]

    container = container_factory(FooService)
    container.start()

    queues_before = list_queues()

    with ServiceRpcClient('foobar') as foo:
        queues_during = list_queues()
        assert foo.spam(ham='eggs') == 'eggs'

    eventlet.sleep(.3)  # sleep for >TTL
    queues_after = list_queues()

    assert queues_before != queues_during
    assert queues_after == queues_before

    # check client re-use
    with ServiceRpcClient('foobar') as foo:
        assert foo.spam(ham='eggs') == 'eggs'
        assert foo.spam(ham='eggs') == 'eggs'

    eventlet.sleep(.3)  # sleep for >TTL
    queues_after = list_queues()
    assert queues_after == queues_before


@pytest.mark.usefixtures('rabbit_config')
@patch('nameko.standalone.rpc.RPC_REPLY_QUEUE_TTL', new=200)
def test_reply_queue_not_removed_while_in_use(
    rabbit_manager, get_vhost, container_factory
):
    def list_queues():
        vhost = get_vhost(config['AMQP_URI'])
        return [
            queue['name']
            for queue in rabbit_manager.get_queues(vhost=vhost)
        ]

    container = container_factory(FooService)
    container.start()

    # check client re-use
    with ServiceRpcClient('foobar') as foo:
        queues_before = list_queues()
        # sleep for 2x TTL
        assert foo.sleep(0.4) == 0.4
        queues_between = list_queues()
        assert foo.spam(ham='eggs') == 'eggs'
        queues_after = list_queues()

    assert queues_before == queues_between == queues_after


@skip_if_no_toxiproxy
class TestDisconnectedWhileWaitingForReply(object):

    @pytest.yield_fixture(autouse=True)
    def fast_reconnects(self):

        @contextmanager
        def establish_connection(self):
            with self.create_connection() as conn:
                conn.ensure_connection(
                    self.on_connection_error,
                    self.connect_max_retries,
                    interval_start=0.1,
                    interval_step=0.1)
                yield conn

        with patch.object(
            Consumer, 'establish_connection', new=establish_connection
        ):
            yield

    @pytest.yield_fixture(autouse=True)
    def fast_expiry(self):
        with patch('nameko.standalone.rpc.RPC_REPLY_QUEUE_TTL', new=100):
            yield

    @pytest.fixture(autouse=True)
    def container(
        self, container_factory, rabbit_manager, rabbit_config, toxiproxy,
        fast_expiry
    ):

        def enable_after_queue_expires():
            eventlet.sleep(1)
            toxiproxy.enable()

        class Service(object):
            name = "service"

            @rpc
            def sleep(self):
                toxiproxy.disable()
                eventlet.spawn_n(enable_after_queue_expires)
                return "OK"

        container = container_factory(Service)
        container.start()

        return container

    @pytest.mark.usefixtures('rabbit_config')
    def test_reply_queue_removed_while_disconnected_with_pending_reply(
        self, container, toxiproxy, toxic_reply_listener
    ):
        with pytest.raises(ReplyQueueExpiredWithPendingReplies):
            with ClusterRpcClient() as client:
                client.service.sleep()


@pytest.mark.usefixtures('rabbit_config')
@patch('nameko.standalone.rpc.RPC_REPLY_QUEUE_TTL', new=200)
def test_async_wait_longer_than_expiry(container_factory):
    """ Replies to async requests will be lost if the reply queue expiry is
    shorter than the period waited before consuming the result.
    """

    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar', timeout=0.6) as foo:
        res = foo.sleep.call_async(0.4)
        eventlet.sleep(0.4)

        with pytest.raises(ReplyQueueExpiredWithPendingReplies):
            res.result()


@pytest.mark.usefixtures('rabbit_config')
@patch('nameko.standalone.rpc.RPC_REPLY_QUEUE_TTL', new=200)
def test_request_longer_than_expiry(container_factory):
    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar') as foo:
        assert foo.sleep(0.4) == 0.4


@pytest.mark.usefixtures('rabbit_config')
@patch('nameko.standalone.rpc.RPC_REPLY_QUEUE_TTL', new=200)
def test_inactive_longer_than_expiry(container_factory):
    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar') as foo:
        eventlet.sleep(0.4)
        assert foo.spam(ham='eggs') == 'eggs'


@pytest.mark.usefixtures("rabbit_config")
def test_unexpected_correlation_id(container_factory):
    container = container_factory(FooService)
    container.start()

    service_rpc_client = ServiceRpcClient("foobar")
    with service_rpc_client as client:

        message = Message(channel=None, properties={
            'reply_to': service_rpc_client.reply_listener.queue.routing_key,
            'correlation_id': 'invalid',
            'content_type': 'application/json'
        })
        amqp_uri = config['AMQP_URI']
        exchange = get_rpc_exchange()

        responder = Responder(amqp_uri, exchange, message)
        with patch('nameko.standalone.rpc._logger', autospec=True) as logger:
            responder.send_response(None, None)
            assert client.spam(ham='eggs') == 'eggs'
            assert logger.debug.call_count == 1


@pytest.mark.usefixtures('rabbit_config')
def test_async_rpc(container_factory):

    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar') as foo:
        rep1 = foo.spam.call_async(ham=1)
        rep2 = foo.spam.call_async(ham=2)
        rep3 = foo.spam.call_async(ham=3)
        rep4 = foo.spam.call_async(ham=4)
        rep5 = foo.spam.call_async(ham=5)
        assert rep2.result() == 2
        assert rep3.result() == 3
        assert rep1.result() == 1
        assert rep4.result() == 4
        assert rep5.result() == 5


@pytest.mark.usefixtures('rabbit_config')
def test_multiple_proxies(container_factory):
    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar') as client1:
        res1 = client1.spam.call_async(ham=1)

        with ServiceRpcClient('foobar') as client2:
            res2 = client2.spam.call_async(ham=2)

            assert res1.result() == 1
            assert res2.result() == 2


@pytest.mark.usefixtures('rabbit_config')
def test_multiple_calls_to_result(container_factory):
    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar') as client:
        res = client.spam.call_async(ham=1)
        res.result()
        res.result()


@skip_if_no_toxiproxy
class TestDisconnectWithPendingReply(object):

    @pytest.yield_fixture
    def toxic_rpc_client(self, rabbit_config, toxiproxy):
        with config.patch({'AMQP_URI': toxiproxy.uri}):
            yield ClusterRpcClient()

    def test_disconnect_and_successfully_reconnect(
        self, container_factory, rabbit_manager, toxic_rpc_client, toxiproxy
    ):
        block = Event()

        class Service(object):
            name = "service"

            @rpc
            def method(self, arg):
                block.wait()
                return arg

        container = container_factory(Service)
        container.start()

        with toxic_rpc_client as client:

            # make an async call that will block,
            # wait for the worker to have spawned
            with wait_for_call(container, 'spawn_worker'):
                res = client.service.method.call_async('msg1')

            # disable toxiproxy to kill connections
            toxiproxy.disable()

            # re-enable toxiproxy when the connection error is detected
            def reconnect(args, kwargs, res, exc_info):
                block.send(True)
                toxiproxy.enable()
                return True

            with wait_for_call(
                toxic_rpc_client.reply_listener.consumer,
                'on_connection_error',
                callback=reconnect
            ):
                # rpc client should recover the message in flight
                res.result() == "msg1"

                # client should work again after reconnection
                assert client.service.method("msg2") == "msg2"

            # stop container before we stop toxiproxy
            container.stop()


@pytest.mark.usefixtures('rabbit_config')
def test_timeout_not_needed(container_factory, rabbit_manager):
    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar', timeout=1) as client:
        assert client.sleep() == 0


@pytest.mark.usefixtures('rabbit_config')
def test_timeout(container_factory, rabbit_manager):
    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar', timeout=.5) as client:
        with pytest.raises(RpcTimeout):
            client.sleep(seconds=2)

        # make sure we can still use the client
        assert client.sleep(seconds=0) == 0


@pytest.mark.usefixtures('rabbit_config')
def test_no_timeout(container_factory, rabbit_manager):
    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar') as client:
        with pytest.raises(eventlet.Timeout):
            with eventlet.Timeout(.1):
                client.sleep(seconds=1)


@pytest.mark.usefixtures('rabbit_config')
def test_async_timeout(container_factory, rabbit_manager):
    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar', timeout=.5) as client:
        result = client.sleep.call_async(seconds=2)
        with pytest.raises(RpcTimeout):
            result.result()


@pytest.mark.usefixtures('rabbit_config')
def test_use_after_close(container_factory, rabbit_manager):
    container = container_factory(FooService)
    container.start()

    with ServiceRpcClient('foobar') as client:
        client.spam(ham=1)
        pass

    with pytest.raises(RuntimeError) as exc:
        client.spam(ham=1)
    assert 'can no longer be used' in str(exc)


@pytest.mark.usefixtures('rabbit_config')
@patch('nameko.standalone.rpc.RPC_REPLY_QUEUE_TTL', new=100)
def test_client_queue_expired_even_if_unused(rabbit_manager, get_vhost):
    vhost = get_vhost(config['AMQP_URI'])
    with ServiceRpcClient('exampleservice'):
        assert len(rabbit_manager.get_queues(vhost)) == 1

    eventlet.sleep(.15)  # sleep for >TTL
    assert len(rabbit_manager.get_queues(vhost)) == 0


@pytest.mark.usefixtures('rabbit_config')
def test_cluster_client(container_factory, rabbit_manager):
    container = container_factory(FooService)
    container.start()

    with ClusterRpcClient() as client:
        assert client.foobar.spam(ham=1) == 1


@pytest.mark.usefixtures('rabbit_config')
def test_cluster_client_reuse(container_factory, rabbit_manager):
    container = container_factory(FooService)
    container.start()

    cluster_client = ClusterRpcClient()
    with cluster_client as client:
        assert client.foobar.spam(ham=1) == 1

    with cluster_client as second_client:
        assert second_client.foobar.spam(ham=1) == 1


@pytest.mark.usefixtures('rabbit_config')
def test_cluster_client_dict_access(container_factory, rabbit_manager):
    container = container_factory(FooService)
    container.start()

    with ClusterRpcClient() as client:
        assert client['foobar'].spam(ham=3) == 3


@pytest.mark.usefixtures('rabbit_config')
def test_recover_from_keyboardinterrupt(container_factory, rabbit_manager):
    container = container_factory(FooService)
    container.start()  # create rpc queues
    container.stop()  # but make sure call doesn't complete

    with ServiceRpcClient('foobar') as client:

        with patch('kombu.connection.Connection.drain_events') as drain_events:
            drain_events.side_effect = KeyboardInterrupt('killing from test')
            with pytest.raises(KeyboardInterrupt):
                client.spam(ham=0)

        container = container_factory(FooService)
        container.start()

        # client should still work
        assert client.spam(ham=1) == 1


class TestConfigurability(object):
    """
    Test and demonstrate configuration options for the standalone RPC client
    """

    @pytest.yield_fixture
    def get_producer(self):
        with patch('nameko.amqp.publish.get_producer') as get_producer:
            yield get_producer

    @pytest.fixture
    def producer(self, get_producer):
        producer = get_producer().__enter__.return_value
        # make sure we don't raise UndeliverableMessage if mandatory is True
        producer.channel.returned_messages.get_nowait.side_effect = queue.Empty
        return producer

    @pytest.mark.usefixtures("memory_rabbit_config")
    @pytest.mark.parametrize("parameter", [
        # delivery options
        'delivery_mode', 'priority', 'expiration',
        # message options
        'serializer', 'compression',
        # retry policy
        'retry', 'retry_policy',
        # other arbitrary publish kwargs
        'user_id', 'bogus_param'
    ])
    def test_regular_parameters(
        self, parameter, mock_container, producer
    ):
        """ Verify that most parameters can be specified at ServiceRpc
        instantiation time.
        """
        value = Mock()

        rpc_client = ClusterRpcClient(**{parameter: value})
        with rpc_client as client:
            client.service.method.call_async()
        assert producer.publish.call_args[1][parameter] == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    @pytest.mark.usefixtures('predictable_call_ids')
    def test_headers(self, mock_container, producer):
        """ Headers can be provided at instantiation time, and are merged with
        Nameko headers.
        """
        data = {'context': 'data'}

        value = {'foo': Mock()}

        rpc_client = ClusterRpcClient(
            context_data=data, **{'headers': value}
        )

        with rpc_client as client:
            client['service-name'].method.call_async()

        def merge_dicts(base, *updates):
            merged = base.copy()
            [merged.update(update) for update in updates]
            return merged

        nameko_headers = {
            'nameko.context': 'data',
            'nameko.call_id_stack': ['standalone_rpc_client.0.0'],
        }

        assert producer.publish.call_args[1]['headers'] == merge_dicts(
            nameko_headers, value
        )

    @pytest.mark.usefixtures('rabbit_config')
    @patch('nameko.rpc.uuid')
    @patch('nameko.standalone.rpc.uuid')
    def test_restricted_parameters(
        self, patch_standalone_uuid, patch_uuid, mock_container, producer
    ):
        """ Verify that providing routing parameters at instantiation
        time has no effect.
        """
        patch_standalone_uuid.uuid4.return_value = "uuid1"
        patch_uuid.uuid4.return_value = "uuid2"

        restricted_params = (
            'exchange', 'routing_key', 'mandatory',
            'correlation_id', 'reply_to'
        )

        rpc_client = ClusterRpcClient(
            **{param: Mock() for param in restricted_params}
        )

        with rpc_client as client:
            client.service.method.call_async()

        publish_params = producer.publish.call_args[1]

        assert publish_params['exchange'].name == "nameko-rpc"
        assert publish_params['routing_key'] == 'service.method'
        assert publish_params['mandatory'] is True
        assert publish_params['reply_to'] == "uuid1"
        assert publish_params['correlation_id'] == "uuid2"


@pytest.mark.filterwarnings("ignore:Mandatory delivery:UserWarning")
@skip_if_no_toxiproxy
class TestStandaloneClientDisconnections(object):

    @pytest.fixture(autouse=True)
    def container(self, container_factory, rabbit_config):

        class Service(object):
            name = "service"

            @rpc
            def echo(self, arg):
                return arg

        container = container_factory(Service)
        container.start()

    @pytest.yield_fixture(autouse=True)
    def retry(self, request):
        retry = False
        if "publish_retry" in request.keywords:
            retry = True

        with patch.object(
            ServiceRpcClient.publisher_cls, 'retry', new=retry
        ):
            yield

    @pytest.yield_fixture(params=[True, False])
    def use_confirms(self, request):
        with patch.object(
            ServiceRpcClient.publisher_cls, 'use_confirms',
            new=request.param
        ):
            yield request.param

    @pytest.yield_fixture
    def service_rpc(self, rabbit_config, toxiproxy):
        with ServiceRpcClient("service", uri=toxiproxy.uri) as client:
            yield client

    @pytest.mark.usefixtures('use_confirms')
    def test_normal(self, service_rpc):
        assert service_rpc.echo(1) == 1
        assert service_rpc.echo(2) == 2

    @pytest.mark.usefixtures('use_confirms')
    def test_down(self, service_rpc, toxiproxy):
        toxiproxy.disable()

        with pytest.raises(OperationalError) as exc_info:
            service_rpc.echo(1)
        assert "ECONNREFUSED" in str(exc_info.value)

    @pytest.mark.usefixtures('use_confirms')
    def test_timeout(self, service_rpc, toxiproxy):
        toxiproxy.set_timeout()

        with pytest.raises(OperationalError):
            service_rpc.echo(1)

    def test_reuse_when_down(self, service_rpc, toxiproxy):
        """ Verify we detect stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        assert service_rpc.echo(1) == 1

        toxiproxy.disable()

        # publisher cannot connect, raises
        with pytest.raises(IOError):
            service_rpc.echo(2)

    def test_reuse_when_recovered(self, service_rpc, toxiproxy):
        """ Verify we detect and recover from stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        assert service_rpc.echo(1) == 1

        toxiproxy.disable()

        # publisher cannot connect, raises
        with pytest.raises(IOError):
            service_rpc.echo(2)

        toxiproxy.enable()

        assert service_rpc.echo(3) == 3

    @pytest.mark.publish_retry
    def test_with_retry_policy(self, service_rpc, toxiproxy):
        """ Verify we automatically recover from stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        assert service_rpc.echo(1) == 1

        toxiproxy.disable()

        def enable_after_retry(args, kwargs, res, exc_info):
            toxiproxy.enable()
            return True

        # subsequent calls succeed (after reconnecting via retry policy)
        with wait_for_call(
            Connection,
            "_establish_connection",
            callback=enable_after_retry
        ):
            assert service_rpc.echo(2) == 2


@skip_if_no_toxiproxy
class TestStandaloneClientReplyListenerDisconnections(object):

    @pytest.yield_fixture(autouse=True)
    def fast_reconnects(self):

        @contextmanager
        def establish_connection(self):
            with self.create_connection() as conn:
                conn.ensure_connection(
                    self.on_connection_error,
                    self.connect_max_retries,
                    interval_start=0.1,
                    interval_step=0.1)
                yield conn

        with patch.object(
            Consumer, 'establish_connection', new=establish_connection
        ):
            yield

    @pytest.fixture(autouse=True)
    def container(self, container_factory, rabbit_config):

        class Service(object):
            name = "service"

            @rpc
            def echo(self, arg):
                return arg

        # very fast heartbeat (2 seconds)
        with config.patch({HEARTBEAT_CONFIG_KEY: 2}):
            container = container_factory(Service)
            container.start()

    @pytest.fixture
    def rpc_client(self, rabbit_config, toxic_reply_listener):
        return ServiceRpcClient('service')

    @pytest.fixture
    def reply_listener(self, rpc_client):
        return rpc_client.reply_listener

    @pytest.yield_fixture
    def service_rpc(self, rpc_client):
        with rpc_client as client:
            yield client

    def test_normal(self, container, service_rpc):
        assert service_rpc.echo("foo") == "foo"

    def test_down(self, reply_listener, service_rpc, toxiproxy):
        """ Verify we detect and recover from closed sockets.

        This failure mode closes the socket between the consumer and the
        rabbit broker.

        Attempting to read from the closed socket raises a socket.error
        and the connection is re-established.
        """
        def reset(args, kwargs, result, exc_info):
            toxiproxy.enable()
            return True

        with wait_for_call(
            reply_listener.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.disable()

            # connection re-established after one error;
            # make the request _inside_ the context block because the reply
            # listener doesn't establish a connection until consuming a result
            assert service_rpc.echo("foo") == "foo"

    def test_upstream_timeout(self, reply_listener, service_rpc, toxiproxy):
        """ Verify we detect and recover from sockets timing out.

        This failure mode means that the socket between the consumer and the
        rabbit broker times out after `timeout` milliseconds and then closes.

        Attempting to read from the socket after it's closed raises a
        socket.error and the connection will be re-established. If `timeout`
        is longer than twice the heartbeat interval, the behaviour is the same
        as in `test_upstream_blackhole` below.
        """
        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            reply_listener.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(timeout=100)

            # connection re-established after one error;
            # make the request _inside_ the context block because the reply
            # listener doesn't establish a connection until consuming a result
            assert service_rpc.echo("foo") == "foo"

    def test_upstream_blackhole(self, reply_listener, service_rpc, toxiproxy):
        """ Verify we detect and recover from sockets losing data.

        This failure mode means that all data sent from the consumer to the
        rabbit broker is lost, but the socket remains open.

        Heartbeats sent from the consumer are not received by the broker. After
        two beats are missed the broker closes the connection, and subsequent
        reads from the socket raise a socket.error, so the connection is
        re-established.
        """
        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            reply_listener.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(timeout=0)

            # connection re-established after one error;
            # make the request _inside_ the context block because the reply
            # listener doesn't establish a connection until consuming a result
            assert service_rpc.echo("foo") == "foo"

    def test_downstream_timeout(self, reply_listener, service_rpc, toxiproxy):
        """ Verify we detect and recover from sockets timing out.

        This failure mode means that the socket between the rabbit broker and
        the consumer times out after `timeout` milliseconds and then closes.

        Attempting to read from the socket after it's closed raises a
        socket.error and the connection will be re-established. If `timeout`
        is longer than twice the heartbeat interval, the behaviour is the same
        as in `test_downstream_blackhole` below, except that the consumer
        cancel will eventually (`timeout` milliseconds) raise a socket.error,
        which is ignored, allowing the teardown to continue.

        See :meth:`kombu.messsaging.Consumer.__exit__`
        """
        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            reply_listener.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(stream="downstream", timeout=100)

            # connection re-established after one error;
            # make the request _inside_ the context block because the reply
            # listener doesn't establish a connection until consuming a result
            assert service_rpc.echo("foo") == "foo"

    def test_downstream_blackhole(
        self, reply_listener, service_rpc, toxiproxy
    ):  # pragma: no cover
        """ Verify we detect and recover from sockets losing data.

        This failure mode means that all data sent from the rabbit broker to
        the consumer is lost, but the socket remains open.

        Heartbeat acknowledgements from the broker are not received by the
        consumer. After two beats are missed the consumer raises a "too many
        heartbeats missed" error.

        Cancelling the consumer requests an acknowledgement from the broker,
        which is swallowed by the socket. There is no timeout when reading
        the acknowledgement so this hangs forever.

        See :meth:`kombu.messsaging.Consumer.__exit__`
        """
        pytest.skip("skip until kombu supports recovery in this scenario")

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with wait_for_call(
            reply_listener.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(stream="downstream", timeout=0)

            # connection re-established after one error;
            # make the request _inside_ the context block because the reply
            # listener doesn't establish a connection until consuming a result
            assert service_rpc.echo("foo") == "foo"


class TestSSL(object):

    @pytest.mark.usefixtures("rabbit_config")
    def test_rpc_client_over_ssl(
        self, container_factory, rabbit_ssl_uri, rabbit_ssl_options
    ):
        class Service(object):
            name = "service"

            @rpc
            def echo(self, *args, **kwargs):
                return args, kwargs

        container = container_factory(Service)
        container.start()

        ssl = rabbit_ssl_options
        uri = rabbit_ssl_uri

        with ServiceRpcClient(
            "service",
            uri=uri,
            ssl=ssl,
            reply_listener_uri=config["AMQP_URI"],
            reply_listener_ssl=False,
        ) as client:
            assert client.echo("a", "b", foo="bar") == [
                ['a', 'b'], {'foo': 'bar'}
            ]



import click
import pytest

from nameko.cli.click_paramtypes import HostPortParamType, KeyValParamType


@pytest.mark.parametrize(
    "value,expected", [
        ("KEY=value", ("KEY", "value")),
        ("FLAG", ("FLAG", True)),
        (3.14, click.exceptions.BadParameter("Value must by a string"))
    ]
)
def test_key_value(value, expected):
    ctype = KeyValParamType()
    if isinstance(expected, Exception):
        with pytest.raises(expected.__class__) as exc_info:
            result = ctype.convert(value, None, None)
        assert exc_info.value.message == expected.message
    else:
        result = ctype.convert(value, None, None)
        assert result == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        ("9999", ("localhost", 9999)),
        ("localhost:9999", ("localhost", 9999)),
        ("host:9999", ("host", 9999)),
        pytest.param(
            "9999a", click.exceptions.BadParameter("9999a is not a valid port number")
        ),
    ],
)
def test_host_port(value, expected):
    ctype = HostPortParamType()
    if isinstance(expected, Exception):
        with pytest.raises(expected.__class__) as exc_info:
            result = ctype.convert(value, None, None)
        assert exc_info.value.message == expected.message
    else:
        result = ctype.convert(value, None, None)
        assert result == expected



from __future__ import unicode_literals

import inspect
import sys

import six


if sys.version_info >= (3, 3):  # pragma: no cover
    from collections.abc import Iterable  # pylint: disable=E0611,E0401
else:  # pragma: no cover
    from collections import Iterable  # pylint: disable=E0611,E0401


class ExtensionNotFound(AttributeError):
    pass


class RpcTimeout(Exception):
    pass


class ReplyQueueExpiredWithPendingReplies(Exception):
    pass


class ContainerBeingKilled(Exception):
    """Raised by :meth:`Container.spawn_worker` if it has started a ``kill``
    sequence.

    Entrypoints should catch this and react as if they hadn't been available
    in the first place, e.g. an rpc consumer should probably requeue
    the message.

    We need this because eventlet may yield during the execution of
    :meth:`Container.kill`, giving entrypoints a chance to fire before
    they themselves have been killed.
    """


registry = {}


def get_module_path(exc_type):
    """ Return the dotted module path of `exc_type`, including the class name.

    e.g.::

        >>> get_module_path(MethodNotFound)
        >>> "nameko.exceptions.MethodNotFound"

    """
    module = inspect.getmodule(exc_type)
    return "{}.{}".format(module.__name__, exc_type.__name__)


class RemoteError(Exception):
    """ Exception to raise at the caller if an exception occurred in the
    remote worker.
    """
    def __init__(self, exc_type=None, value=""):
        self.exc_type = exc_type
        self.value = value
        message = '{} {}'.format(exc_type, value)
        super(RemoteError, self).__init__(message)


def safe_for_serialization(value):
    """ Transform a value in preparation for serializing as json

    no-op for strings, mappings and iterables have their entries made safe,
    and all other values are stringified, with a fallback value if that fails
    """

    if isinstance(value, six.string_types):
        return value
    if isinstance(value, dict):
        return {
            safe_for_serialization(key): safe_for_serialization(val)
            for key, val in six.iteritems(value)
        }
    if isinstance(value, Iterable):
        return list(map(safe_for_serialization, value))

    try:
        return six.text_type(value)
    except Exception:
        return '[__unicode__ failed]'


def serialize(exc):
    """ Serialize `self.exc` into a data dictionary representing it.
    """

    return {
        'exc_type': type(exc).__name__,
        'exc_path': get_module_path(type(exc)),
        'exc_args': list(map(safe_for_serialization, exc.args)),
        'value': safe_for_serialization(exc),
    }


def deserialize(data):
    """ Deserialize `data` to an exception instance.

    If the `exc_path` value matches an exception registered as
    ``deserializable``, return an instance of that exception type.
    Otherwise, return a `RemoteError` instance describing the exception
    that occurred.
    """
    key = data.get('exc_path')
    if key in registry:
        exc_args = data.get('exc_args', ())
        return registry[key](*exc_args)

    exc_type = data.get('exc_type')
    value = data.get('value')
    return RemoteError(exc_type=exc_type, value=value)


def deserialize_to_instance(exc_type):
    """ Decorator that registers `exc_type` as deserializable back into an
    instance, rather than a :class:`RemoteError`. See :func:`deserialize`.
    """
    key = get_module_path(exc_type)
    registry[key] = exc_type
    return exc_type


class BadRequest(Exception):
    pass


@deserialize_to_instance
class MalformedRequest(BadRequest):
    pass


@deserialize_to_instance
class MethodNotFound(BadRequest):
    pass


@deserialize_to_instance
class IncorrectSignature(BadRequest):
    pass


class UnknownService(Exception):
    def __init__(self, service_name):
        self._service_name = service_name
        super(UnknownService, self).__init__(service_name)

    def __str__(self):
        return "Unknown service `{}`".format(self._service_name)


class UnserializableValueError(Exception):
    def __init__(self, value):
        try:
            self.repr_value = repr(value)
        except Exception:
            self.repr_value = '[__repr__ failed]'
        super(UnserializableValueError, self).__init__()

    def __str__(self):
        return "Unserializable value: `{}`".format(self.repr_value)


class ConfigurationError(Exception):
    pass


class CommandError(Exception):
    """Raise from subcommands to report error back to the user"""


class ConnectionNotFound(BadRequest):
    """Unknown websocket connection id"""



from __future__ import absolute_import

from contextlib import contextmanager

import eventlet
import pytest
from eventlet.event import Event
from eventlet.semaphore import Semaphore
from kombu import Exchange, Queue
from kombu.connection import Connection
from kombu.exceptions import OperationalError
from mock import Mock, call, patch
from six.moves import queue

from nameko import config
from nameko.amqp.consume import Consumer as ConsumerCore
from nameko.amqp.publish import Publisher as PublisherCore
from nameko.amqp.publish import get_producer
from nameko.constants import AMQP_URI_CONFIG_KEY, HEARTBEAT_CONFIG_KEY
from nameko.containers import WorkerContext
from nameko.exceptions import ContainerBeingKilled
from nameko.messaging import (
    Consumer, Publisher, consume, decode_from_headers, encode_to_headers
)
from nameko.testing.services import dummy, entrypoint_hook, entrypoint_waiter
from nameko.testing.utils import (
    ANY_PARTIAL, DummyProvider, get_extension, wait_for_call
)
from nameko.testing.waiting import wait_for_call as patch_wait
from nameko.utils.retry import retry

from test import skip_if_no_toxiproxy


foobar_ex = Exchange('foobar_ex', durable=False)
foobar_queue = Queue('foobar_queue', exchange=foobar_ex, durable=False)

CONSUME_TIMEOUT = 1.2  # a bit more than 1 second


@pytest.yield_fixture
def patch_maybe_declare():
    with patch('nameko.messaging.maybe_declare', autospec=True) as patched:
        yield patched


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.usefixtures("predictable_call_ids")
def test_publish_to_exchange(
    patch_maybe_declare, mock_channel, mock_producer, mock_container
):
    container = mock_container
    container.service_name = "srcservice"

    service = Mock()
    worker_ctx = WorkerContext(container, service, DummyProvider("publish"))

    publisher = Publisher(exchange=foobar_ex).bind(container, "publish")

    # test declarations
    publisher.setup()
    assert patch_maybe_declare.call_args_list == [
        call(foobar_ex, mock_channel)
    ]

    # test publish
    msg = "msg"
    service.publish = publisher.get_dependency(worker_ctx)
    service.publish(msg, publish_kwarg="value")

    headers = {
        'nameko.call_id_stack': ['srcservice.publish.0']
    }
    expected_args = ('msg',)
    expected_kwargs = {
        'publish_kwarg': "value",
        'exchange': foobar_ex,
        'headers': headers,
        'declare': publisher.declare,
        'retry': publisher.publisher_cls.retry,
        'retry_policy': publisher.publisher_cls.retry_policy,
        'compression': publisher.publisher_cls.compression,
        'mandatory': publisher.publisher_cls.mandatory,
        'expiration': publisher.publisher_cls.expiration,
        'delivery_mode': publisher.publisher_cls.delivery_mode,
        'priority': publisher.publisher_cls.priority,
        'serializer': publisher.publisher_cls.serializer
    }

    assert mock_producer.publish.call_args_list == [
        call(*expected_args, **expected_kwargs)
    ]


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.usefixtures("predictable_call_ids")
def test_publish_custom_headers(mock_container, mock_producer):
    container = mock_container
    container.service_name = "srcservice"

    ctx_data = {'language': 'en', 'customheader': 'customvalue'}
    service = Mock()
    worker_ctx = WorkerContext(
        container, service, DummyProvider('method'), data=ctx_data
    )

    publisher = Publisher(
        exchange=foobar_ex, declare=[foobar_queue]).bind(container, "publish")
    publisher.setup()

    msg = "msg"
    service.publish = publisher.get_dependency(worker_ctx)

    # context data changes are reflected up to the point of publishing
    worker_ctx.data['language'] = 'fr'

    service.publish(msg, publish_kwarg="value")

    expected_args = ('msg',)
    expected_headers = {
        'nameko.language': 'fr',
        'nameko.customheader': 'customvalue',
        'nameko.call_id_stack': ['srcservice.method.0'],
    }
    expected_kwargs = {
        'publish_kwarg': "value",
        'exchange': foobar_ex,
        'headers': expected_headers,
        'declare': publisher.declare,
        'retry': publisher.publisher_cls.retry,
        'retry_policy': publisher.publisher_cls.retry_policy,
        'compression': publisher.publisher_cls.compression,
        'mandatory': publisher.publisher_cls.mandatory,
        'expiration': publisher.publisher_cls.expiration,
        'delivery_mode': publisher.publisher_cls.delivery_mode,
        'priority': publisher.publisher_cls.priority,
        'serializer': publisher.publisher_cls.serializer
    }

    assert mock_producer.publish.call_args_list == [
        call(*expected_args, **expected_kwargs)
    ]


@pytest.mark.usefixtures('empty_config')
def test_encode_headers():

    context_data = {
        'foo': 'FOO',
        'bar': 'BAR',
        'baz': 'BAZ',
        # unserialisable, shouldn't be included in the processed headers
        'none': None,
    }

    res = encode_to_headers(context_data, prefix="testprefix")
    assert res == {
        'testprefix.foo': 'FOO',
        'testprefix.bar': 'BAR',
        'testprefix.baz': 'BAZ',
    }


def test_decode_headers():

    headers = {
        'testprefix.foo': 'FOO',
        'testprefix.bar': 'BAR',
        'testprefix.baz': 'BAZ',
        'differentprefix.foo': 'XXX',
        'testprefix.call_id_stack': ['a', 'b', 'c'],
    }

    res = decode_from_headers(headers, prefix="testprefix")
    assert res == {
        'foo': 'FOO',
        'bar': 'BAR',
        'baz': 'BAZ',
        'call_id_stack': ['a', 'b', 'c'],
        'differentprefix.foo': 'XXX'
    }


# =============================================================================
# INTEGRATION TESTS
# =============================================================================

@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.usefixtures("predictable_call_ids")
def test_publish_to_rabbit(rabbit_manager, get_vhost, mock_container):

    vhost = get_vhost(config['AMQP_URI'])

    container = mock_container
    container.service_name = "service"

    ctx_data = {'language': 'en', 'customheader': 'customvalue'}
    service = Mock()
    worker_ctx = WorkerContext(
        container, service, DummyProvider('method'), data=ctx_data
    )

    publisher = Publisher(
        exchange=foobar_ex, declare=[foobar_queue]
    ).bind(container, "publish")

    publisher.setup()
    publisher.start()

    # test queue, exchange and binding created in rabbit
    exchanges = rabbit_manager.get_exchanges(vhost)
    queues = rabbit_manager.get_queues(vhost)
    bindings = rabbit_manager.get_queue_bindings(vhost, foobar_queue.name)

    assert "foobar_ex" in [exchange['name'] for exchange in exchanges]
    assert "foobar_queue" in [queue['name'] for queue in queues]
    assert "foobar_ex" in [binding['source'] for binding in bindings]

    service.publish = publisher.get_dependency(worker_ctx)
    service.publish("msg")

    # test message published to queue
    messages = rabbit_manager.get_messages(vhost, foobar_queue.name)
    assert ['"msg"'] == [msg['payload'] for msg in messages]

    # test message headers
    assert messages[0]['properties']['headers'] == {
        'nameko.language': 'en',
        'nameko.customheader': 'customvalue',
        'nameko.call_id_stack': ['service.method.0'],
    }


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.usefixtures("predictable_call_ids")
def test_unserialisable_headers(rabbit_manager, get_vhost, mock_container):

    vhost = get_vhost(config['AMQP_URI'])

    container = mock_container
    container.service_name = "service"
    container.spawn_managed_thread = eventlet.spawn

    ctx_data = {'language': 'en', 'customheader': None}
    service = Mock()
    worker_ctx = WorkerContext(
        container, service, DummyProvider('method'), data=ctx_data
    )

    publisher = Publisher(
        exchange=foobar_ex, declare=[foobar_queue]).bind(container, "publish")

    publisher.setup()
    publisher.start()

    service.publish = publisher.get_dependency(worker_ctx)
    service.publish("msg")
    messages = rabbit_manager.get_messages(vhost, foobar_queue.name)

    assert messages[0]['properties']['headers'] == {
        'nameko.language': 'en',
        'nameko.call_id_stack': ['service.method.0'],
        # no `customheader`
    }


@pytest.mark.usefixtures("rabbit_config")
def test_consume_from_rabbit(get_vhost, rabbit_manager, mock_container):

    vhost = get_vhost(config['AMQP_URI'])

    container = mock_container
    container.shared_extensions = {}
    container.worker_ctx_cls = WorkerContext
    container.service_name = "service"
    container.max_workers = 10

    content_type = 'application/data'
    container.accept = [content_type]

    def spawn_managed_thread(method, identifier=None):
        return eventlet.spawn(method)

    container.spawn_managed_thread = spawn_managed_thread

    worker_ctx = WorkerContext(container, None, DummyProvider())

    consumer = Consumer(
        queue=foobar_queue, requeue_on_error=False).bind(container, "publish")

    consumer.setup()
    consumer.start()

    # test queue, exchange and binding created in rabbit
    exchanges = rabbit_manager.get_exchanges(vhost)
    queues = rabbit_manager.get_queues(vhost)
    bindings = rabbit_manager.get_queue_bindings(vhost, foobar_queue.name)

    assert "foobar_ex" in [exchange['name'] for exchange in exchanges]
    assert "foobar_queue" in [queue['name'] for queue in queues]
    assert "foobar_ex" in [binding['source'] for binding in bindings]

    # test message consumed from queue
    container.spawn_worker.return_value = worker_ctx

    headers = {'nameko.language': 'en', 'nameko.customheader': 'customvalue'}

    rabbit_manager.publish(
        vhost, foobar_ex.name, '', 'msg',
        properties=dict(headers=headers, content_type=content_type))

    ctx_data = {
        'language': 'en',
        'customheader': 'customvalue',
    }
    with wait_for_call(CONSUME_TIMEOUT, container.spawn_worker) as method:
        method.assert_called_once_with(consumer, ('msg',), {},
                                       context_data=ctx_data,
                                       handle_result=ANY_PARTIAL)
        handle_result = method.call_args[1]['handle_result']

    # ack message
    handle_result(worker_ctx, 'result')

    consumer.stop()


@skip_if_no_toxiproxy
class TestConsumerDisconnections(object):
    """ Test and demonstrate behaviour under poor network conditions.
    """
    @pytest.yield_fixture(autouse=True)
    def fast_reconnects(self):

        @contextmanager
        def establish_connection(self):
            with self.create_connection() as conn:
                conn.ensure_connection(
                    self.on_connection_error,
                    self.connect_max_retries,
                    interval_start=0.1,
                    interval_step=0.1)
                yield conn

        with patch.object(
            ConsumerCore, 'establish_connection', new=establish_connection
        ):
            yield

    @pytest.yield_fixture
    def toxic_consumer(self, toxiproxy):
        with patch.object(Consumer, 'amqp_uri', new=toxiproxy.uri):
            yield

    @pytest.fixture
    def queue(self,):
        queue = Queue(name="queue")
        return queue

    @pytest.fixture
    def publish(self, rabbit_config, queue):
        amqp_uri = config[AMQP_URI_CONFIG_KEY]

        def publish(msg):
            with get_producer(amqp_uri) as producer:
                producer.publish(
                    msg,
                    serializer="json",
                    routing_key=queue.name
                )
        return publish

    @pytest.fixture
    def lock(self):
        return Semaphore()

    @pytest.fixture
    def tracker(self):
        return Mock()

    @pytest.fixture(autouse=True)
    def container(
        self, container_factory, rabbit_config, toxic_consumer, queue,
        lock, tracker
    ):

        class Service(object):
            name = "service"

            @consume(queue)
            def echo(self, arg):
                lock.acquire()
                lock.release()
                tracker(arg)
                return arg

        # very fast heartbeat (2 seconds)
        with config.patch({HEARTBEAT_CONFIG_KEY: 2}):
            container = container_factory(Service)
            container.start()

            # we have to let the container connect before disconnecting
            # otherwise we end up in retry_over_time trying to make the
            # initial connection; we get stuck there because it has a
            # function-local copy of "on_connection_error" that is never patched
            eventlet.sleep(.05)

            yield container

    def test_normal(self, container, publish):
        msg = "foo"
        with entrypoint_waiter(container, 'echo') as result:
            publish(msg)
        assert result.get() == msg

    def test_down(self, container, publish, toxiproxy):
        """ Verify we detect and recover from closed sockets.

        This failure mode closes the socket between the consumer and the
        rabbit broker.

        Attempting to read from the closed socket raises a socket.error
        and the connection is re-established.
        """
        consumer = get_extension(container, Consumer)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.enable()
            return True

        with patch_wait(
            consumer.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.disable()

        # connection re-established
        msg = "foo"
        with entrypoint_waiter(container, 'echo') as result:
            publish(msg)
        assert result.get() == msg

    def test_upstream_timeout(self, container, publish, toxiproxy):
        """ Verify we detect and recover from sockets timing out.

        This failure mode means that the socket between the consumer and the
        rabbit broker times out after `timeout` milliseconds and then closes.

        Attempting to read from the socket after it's closed raises a
        socket.error and the connection will be re-established. If `timeout`
        is longer than twice the heartbeat interval, the behaviour is the same
        as in `test_upstream_blackhole` below.
        """
        consumer = get_extension(container, Consumer)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with patch_wait(
            consumer.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(timeout=100)

        # connection re-established
        msg = "foo"
        with entrypoint_waiter(container, 'echo') as result:
            publish(msg)
        assert result.get() == msg

    def test_upstream_blackhole(self, container, publish, toxiproxy):
        """ Verify we detect and recover from sockets losing data.

        This failure mode means that all data sent from the consumer to the
        rabbit broker is lost, but the socket remains open.

        Heartbeats sent from the consumer are not received by the broker. After
        two beats are missed the broker closes the connection, and subsequent
        reads from the socket raise a socket.error, so the connection is
        re-established.
        """
        consumer = get_extension(container, Consumer)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with patch_wait(
            consumer.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(timeout=0)

        # connection re-established
        msg = "foo"
        with entrypoint_waiter(container, 'echo') as result:
            publish(msg)
        assert result.get() == msg

    def test_downstream_timeout(self, container, publish, toxiproxy):
        """ Verify we detect and recover from sockets timing out.

        This failure mode means that the socket between the rabbit broker and
        the consumer times out after `timeout` milliseconds and then closes.

        Attempting to read from the socket after it's closed raises a
        socket.error and the connection will be re-established. If `timeout`
        is longer than twice the heartbeat interval, the behaviour is the same
        as in `test_downstream_blackhole` below, except that the consumer
        cancel will eventually (`timeout` milliseconds) raise a socket.error,
        which is ignored, allowing the teardown to continue.

        See :meth:`kombu.messsaging.Consumer.__exit__`
        """
        consumer = get_extension(container, Consumer)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with patch_wait(
            consumer.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(stream="downstream", timeout=100)

        # connection re-established
        msg = "foo"
        with entrypoint_waiter(container, 'echo') as result:
            publish(msg)
        assert result.get() == msg

    def test_downstream_blackhole(
        self, container, publish, toxiproxy
    ):  # pragma: no cover
        """ Verify we detect and recover from sockets losing data.

        This failure mode means that all data sent from the rabbit broker to
        the consumer is lost, but the socket remains open.

        Heartbeat acknowledgements from the broker are not received by the
        consumer. After two beats are missed the consumer raises a "too many
        heartbeats missed" error.

        Cancelling the consumer requests an acknowledgement from the broker,
        which is swallowed by the socket. There is no timeout when reading
        the acknowledgement so this hangs forever.

        See :meth:`kombu.messsaging.Consumer.__exit__`
        """
        pytest.skip("skip until kombu supports recovery in this scenario")

        consumer = get_extension(container, Consumer)

        def reset(args, kwargs, result, exc_info):
            toxiproxy.reset_timeout()
            return True

        with patch_wait(
            consumer.consumer, 'on_connection_error', callback=reset
        ):
            toxiproxy.set_timeout(stream="downstream", timeout=0)

        # connection re-established
        msg = "foo"
        with entrypoint_waiter(container, 'echo') as result:
            publish(msg)
        assert result.get() == msg

    def test_message_ack_regression(
        self, container, publish, toxiproxy, lock, tracker
    ):
        """ Regression for https://github.com/nameko/nameko/issues/511
        """
        # prevent workers from completing
        lock.acquire()

        # fire entrypoint and block the worker;
        # break connection while the worker is active, then release worker
        with entrypoint_waiter(container, 'echo') as result:
            publish('msg1')
            while not lock._waiters:
                eventlet.sleep()  # pragma: no cover
            toxiproxy.disable()
            # allow connection to close before releasing worker
            eventlet.sleep(.1)
            lock.release()

        # entrypoint will return and attempt to ack initiating message
        assert result.get() == "msg1"

        # enabling connection will re-deliver the initiating message
        # and it will be processed again
        with entrypoint_waiter(container, 'echo') as result:
            toxiproxy.enable()
        assert result.get() == "msg1"

        # connection re-established, container should work again
        with entrypoint_waiter(container, 'echo', timeout=1) as result:
            publish('msg2')
        assert result.get() == 'msg2'

    def test_message_requeue_regression(
        self, container, publish, toxiproxy, lock, tracker
    ):
        """ Regression for https://github.com/nameko/nameko/issues/511
        """
        # turn on requeue_on_error
        consumer = get_extension(container, Consumer)
        consumer.requeue_on_error = True

        # make entrypoint raise the first time it's called so that
        # we attempt to requeue it
        class Boom(Exception):
            pass

        def error_once():
            yield Boom("error")
            while True:
                yield
        tracker.side_effect = error_once()

        # prevent workers from completing
        lock.acquire()

        # fire entrypoint and block the worker;
        # break connection while the worker is active, then release worker
        with entrypoint_waiter(container, 'echo') as result:
            publish('msg1')
            while not lock._waiters:
                eventlet.sleep()  # pragma: no cover
            toxiproxy.disable()
            # allow connection to close before releasing worker
            eventlet.sleep(.1)
            lock.release()

        # entrypoint will return and attempt to requeue initiating message
        with pytest.raises(Boom):
            result.get()

        # enabling connection will re-deliver the initiating message
        # and it will be processed again
        with entrypoint_waiter(container, 'echo', timeout=1) as result:
            toxiproxy.enable()
        assert result.get() == 'msg1'

        # connection re-established, container should work again
        with entrypoint_waiter(container, 'echo', timeout=1) as result:
            publish('msg2')
        assert result.get() == 'msg2'


@skip_if_no_toxiproxy
class TestPublisherDisconnections(object):
    """ Test and demonstrate behaviour under poor network conditions.

    Publisher confirms must be enabled for some of these tests to pass. Without
    confirms, previously used but now dead connections will accept writes
    without raising. These tests are skipped in this scenario.

    Note that publisher confirms do not protect against sockets that remain
    open but do not deliver messages (i.e. `toxiproxy.set_timeout(0)`).
    This can only be mitigated with AMQP heartbeats (not yet supported)
    """

    @pytest.fixture
    def tracker(self):
        return Mock()

    @pytest.yield_fixture(params=[True, False])
    def use_confirms(self, request):
        with patch.object(
            Publisher.publisher_cls, 'use_confirms', new=request.param
        ):
            yield request.param

    @pytest.yield_fixture
    def publisher_container(
        self, request, container_factory, tracker, rabbit_config, toxiproxy
    ):
        retry = False
        if "publish_retry" in request.keywords:
            retry = True

        class Service(object):
            name = "publish"

            publish = Publisher(uri=toxiproxy.uri)

            @dummy
            def send(self, payload):
                tracker("send", payload)
                self.publish(payload, routing_key="test_queue", retry=retry)

        container = container_factory(Service)
        container.start()
        yield container

    @pytest.yield_fixture
    def consumer_container(
        self, container_factory, tracker, rabbit_config
    ):
        class Service(object):
            name = "consume"

            @consume(Queue("test_queue"))
            def recv(self, payload):
                tracker("recv", payload)

        container = container_factory(Service)
        container.start()
        yield container

    @pytest.mark.usefixtures('use_confirms')
    def test_normal(self, publisher_container, consumer_container, tracker):

        # call 1 succeeds
        payload1 = "payload1"
        with entrypoint_waiter(consumer_container, 'recv'):
            with entrypoint_hook(publisher_container, 'send') as send:
                send(payload1)

        assert tracker.call_args_list == [
            call("send", payload1),
            call("recv", payload1),
        ]

        # call 2 succeeds
        payload2 = "payload2"
        with entrypoint_waiter(consumer_container, 'recv'):
            with entrypoint_hook(publisher_container, 'send') as send:
                send(payload2)

        assert tracker.call_args_list == [
            call("send", payload1),
            call("recv", payload1),
            call("send", payload2),
            call("recv", payload2),
        ]

    @pytest.mark.usefixtures('use_confirms')
    def test_down(
        self, publisher_container, consumer_container, tracker, toxiproxy
    ):
        with toxiproxy.disabled():

            payload1 = "payload1"
            with pytest.raises(OperationalError) as exc_info:
                with entrypoint_hook(publisher_container, 'send') as send:
                    send(payload1)
            assert "ECONNREFUSED" in str(exc_info.value)

            assert tracker.call_args_list == [
                call("send", payload1),
            ]

    @pytest.mark.usefixtures('use_confirms')
    def test_timeout(
        self, publisher_container, consumer_container, tracker, toxiproxy
    ):
        with toxiproxy.timeout(500):

            payload1 = "payload1"
            with pytest.raises(OperationalError):  # socket closed
                with entrypoint_hook(publisher_container, 'send') as send:
                    send(payload1)

            assert tracker.call_args_list == [
                call("send", payload1),
            ]

    def test_reuse_when_down(
        self, publisher_container, consumer_container, tracker, toxiproxy,
    ):
        """ Verify we detect stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        # call 1 succeeds
        payload1 = "payload1"
        with entrypoint_waiter(consumer_container, 'recv'):
            with entrypoint_hook(publisher_container, 'send') as send:
                send(payload1)

        assert tracker.call_args_list == [
            call("send", payload1),
            call("recv", payload1),
        ]

        with toxiproxy.disabled():

            # call 2 fails
            payload2 = "payload2"
            with pytest.raises(IOError):
                with entrypoint_hook(publisher_container, 'send') as send:
                    send(payload2)

            assert tracker.call_args_list == [
                call("send", payload1),
                call("recv", payload1),
                call("send", payload2),
            ]

    def test_reuse_when_recovered(
        self, publisher_container, consumer_container, tracker, toxiproxy
    ):
        """ Verify we detect and recover from stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        # call 1 succeeds
        payload1 = "payload1"
        with entrypoint_waiter(consumer_container, 'recv'):
            with entrypoint_hook(publisher_container, 'send') as send:
                send(payload1)

        assert tracker.call_args_list == [
            call("send", payload1),
            call("recv", payload1),
        ]

        with toxiproxy.disabled():

            # call 2 fails
            payload2 = "payload2"
            with pytest.raises(IOError):
                with entrypoint_hook(publisher_container, 'send') as send:
                    send(payload2)

            assert tracker.call_args_list == [
                call("send", payload1),
                call("recv", payload1),
                call("send", payload2),
            ]

        # call 3 succeeds
        payload3 = "payload3"
        with entrypoint_waiter(consumer_container, 'recv'):
            with entrypoint_hook(publisher_container, 'send') as send:
                send(payload3)

        assert tracker.call_args_list == [
            call("send", payload1),
            call("recv", payload1),
            call("send", payload2),
            call("send", payload3),
            call("recv", payload3),
        ]

    @pytest.mark.publish_retry
    def test_with_retry_policy(
        self, publisher_container, consumer_container, tracker, toxiproxy
    ):
        """ Verify we automatically recover from stale connections.

        Publish confirms are required for this functionality. Without confirms
        the later messages are silently lost and the test hangs waiting for a
        response.
        """
        # call 1 succeeds
        payload1 = "payload1"
        with entrypoint_waiter(consumer_container, 'recv'):
            with entrypoint_hook(publisher_container, 'send') as send:
                send(payload1)

        assert tracker.call_args_list == [
            call("send", payload1),
            call("recv", payload1),
        ]

        toxiproxy.disable()

        def enable_after_retry(args, kwargs, res, exc_info):
            toxiproxy.enable()
            return True

        # call 2 succeeds (after reconnecting via retry policy)
        with patch_wait(
            Connection,
            "_establish_connection",
            callback=enable_after_retry
        ):
            payload2 = "payload2"
            with entrypoint_waiter(consumer_container, 'recv'):
                with entrypoint_hook(publisher_container, 'send') as send:
                    send(payload2)

            assert tracker.call_args_list == [
                call("send", payload1),
                call("recv", payload1),
                call("send", payload2),
                call("recv", payload2),
            ]


class TestConsumerConfigurability(object):
    """
    Test and demonstrate configuration options for the Consumer
    """

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_heartbeat(self, mock_container):
        mock_container.service_name = "service"

        value = 999
        queue = Mock()

        consumer = Consumer(
            queue, heartbeat=value
        ).bind(mock_container, "method")
        consumer.setup()

        assert consumer.consumer.connection.heartbeat == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_prefetch_count(self, mock_container):
        mock_container.service_name = "service"

        value = 999
        queue = Mock()

        consumer = Consumer(
            queue, prefetch_count=value
        ).bind(mock_container, "method")
        consumer.setup()

        assert consumer.consumer.prefetch_count == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_accept(self, mock_container):
        mock_container.service_name = "service"

        value = ['yaml', 'json']
        queue = Mock()

        consumer = Consumer(
            queue, accept=value
        ).bind(mock_container, "method")
        consumer.setup()

        assert consumer.consumer.accept == value


class TestPublisherConfigurability(object):
    """
    Test and demonstrate configuration options for the Publisher
    """

    @pytest.yield_fixture
    def get_producer(self):
        with patch('nameko.amqp.publish.get_producer') as get_producer:
            yield get_producer

    @pytest.fixture
    def producer(self, get_producer):
        producer = get_producer().__enter__.return_value
        # make sure we don't raise UndeliverableMessage if mandatory is True
        producer.channel.returned_messages.get_nowait.side_effect = queue.Empty
        return producer

    @pytest.mark.usefixtures("memory_rabbit_config")
    @pytest.mark.parametrize("parameter", [
        # routing
        'exchange', 'routing_key',
        # delivery options
        'delivery_mode', 'mandatory', 'priority', 'expiration',
        # message options
        'serializer', 'compression',
        # retry policy
        'retry', 'retry_policy',
        # other arbitrary publish kwargs
        'correlation_id', 'user_id', 'bogus_param'
    ])
    def test_regular_parameters(
        self, parameter, mock_container, producer
    ):
        """ Verify that most parameters can be specified at instantiation time,
        and overridden at publish time.
        """
        mock_container.service_name = "service"

        worker_ctx = Mock()
        worker_ctx.context_data = {}

        instantiation_value = Mock()
        publish_value = Mock()

        publisher = Publisher(
            **{parameter: instantiation_value}
        ).bind(mock_container, "publish")
        publisher.setup()

        publish = publisher.get_dependency(worker_ctx)

        publish("payload")
        assert producer.publish.call_args[1][parameter] == instantiation_value

        publish("payload", **{parameter: publish_value})
        assert producer.publish.call_args[1][parameter] == publish_value

    @pytest.mark.usefixtures("memory_rabbit_config")
    @pytest.mark.usefixtures('predictable_call_ids')
    def test_headers(self, mock_container, producer):
        """ Headers provided at publish time are merged with any provided
        at instantiation time. Nameko headers are always present.
        """
        mock_container.service_name = "service"

        # use a real worker context so nameko headers are generated
        service = Mock()
        entrypoint = Mock(method_name="method")
        worker_ctx = WorkerContext(
            mock_container, service, entrypoint, data={'context': 'data'}
        )

        nameko_headers = {
            'nameko.context': 'data',
            'nameko.call_id_stack': ['service.method.0'],
        }

        instantiation_value = {'foo': Mock()}
        publish_value = {'bar': Mock()}

        publisher = Publisher(
            **{'headers': instantiation_value}
        ).bind(mock_container, "publish")
        publisher.setup()

        publish = publisher.get_dependency(worker_ctx)

        def merge_dicts(base, *updates):
            merged = base.copy()
            [merged.update(update) for update in updates]
            return merged

        publish("payload")
        assert producer.publish.call_args[1]['headers'] == merge_dicts(
            nameko_headers, instantiation_value
        )

        publish("payload", headers=publish_value)
        assert producer.publish.call_args[1]['headers'] == merge_dicts(
            nameko_headers, instantiation_value, publish_value
        )

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_declare(self, mock_container, producer):
        """ Declarations provided at publish time are merged with any provided
        at instantiation time. Any provided exchange and queue are always
        declared.
        """
        mock_container.service_name = "service"

        worker_ctx = Mock()
        worker_ctx.context_data = {}

        exchange = Mock()

        instantiation_value = [Mock()]
        publish_value = [Mock()]

        publisher = Publisher(
            exchange=exchange, **{'declare': instantiation_value}
        ).bind(mock_container, "publish")
        publisher.setup()

        publish = publisher.get_dependency(worker_ctx)

        publish("payload")
        assert producer.publish.call_args[1]['declare'] == (
            instantiation_value + [exchange]
        )

        publish("payload", declare=publish_value)
        assert producer.publish.call_args[1]['declare'] == (
            instantiation_value + [exchange] + publish_value
        )

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_use_confirms(self, mock_container, get_producer):
        """ Verify that publish-confirms can be set as a default specified at
        instantiation time, which can be overridden by a value specified at
        publish time.
        """
        mock_container.service_name = "service"

        worker_ctx = Mock()
        worker_ctx.context_data = {}

        publisher = Publisher(
            use_confirms=False
        ).bind(mock_container, "publish")
        publisher.setup()

        publish = publisher.get_dependency(worker_ctx)

        publish("payload")
        use_confirms = get_producer.call_args[0][3].get('confirm_publish')
        assert use_confirms is False

        publish("payload", use_confirms=True)
        use_confirms = get_producer.call_args[0][3].get('confirm_publish')
        assert use_confirms is True


class TestPrefetchCount(object):

    @pytest.mark.usefixtures("rabbit_config")
    @config.patch({'PREFETCH_COUNT': 1})
    def test_prefetch_count(
        self, rabbit_manager, get_vhost, container_factory
    ):

        vhost = get_vhost(config['AMQP_URI'])

        messages = []

        exchange = Exchange('exchange')
        queue = Queue('queue', exchange=exchange, auto_delete=False)

        class LimitedConsumer1(Consumer):

            def handle_message(self, body, message):
                consumer_continue.wait()
                super(LimitedConsumer1, self).handle_message(body, message)

        class LimitedConsumer2(Consumer):

            def handle_message(self, body, message):
                messages.append(body)
                super(LimitedConsumer2, self).handle_message(body, message)

        class Service(object):
            name = "service"

            @LimitedConsumer1.decorator(queue=queue)
            @LimitedConsumer2.decorator(queue=queue)
            def handle(self, payload):
                pass

        container = container_factory(Service)
        container.start()

        consumer_continue = Event()

        # the two handlers would ordinarily take alternating messages, but are
        # limited to holding one un-ACKed message. Since Handler1 never ACKs,
        # it only ever gets one message, and Handler2 gets the others.

        def wait_for_expected(worker_ctx, res, exc_info):
            return {'m3', 'm4', 'm5'}.issubset(set(messages))

        with entrypoint_waiter(
            container, 'handle', callback=wait_for_expected
        ):
            properties = {'content_type': 'application/data'}
            for message in ('m1', 'm2', 'm3', 'm4', 'm5'):
                rabbit_manager.publish(
                    vhost, 'exchange', '', message, properties=properties
                )

        # we don't know which handler picked up the first message,
        # but all the others should've been handled by Handler2
        assert messages[-3:] == ['m3', 'm4', 'm5']

        # release the waiting consumer
        consumer_continue.send(None)


class TestContainerBeingKilled(object):

    @pytest.fixture
    def publisher(self, amqp_uri):
        return PublisherCore(amqp_uri)

    @pytest.mark.usefixtures("rabbit_config")
    def test_container_killed(self, container_factory, publisher):
        queue = Queue('queue')

        class Service(object):
            name = "service"

            @consume(queue)
            def method(self, payload):
                pass  # pragma: no cover

        container = container_factory(Service)
        container.start()

        # check message is requeued if container throws ContainerBeingKilled
        with patch.object(container, 'spawn_worker') as spawn_worker:
            spawn_worker.side_effect = ContainerBeingKilled()

            with patch_wait(ConsumerCore, 'requeue_message'):
                publisher.publish("payload", routing_key=queue.name)


class TestSSL(object):

    @pytest.fixture
    def queue(self,):
        queue = Queue(name="queue")
        return queue

    @pytest.fixture(params=[True, False])
    def rabbit_ssl_options(self, request, rabbit_ssl_options):
        verify_certs = request.param
        if verify_certs is False:
            # remove certificate paths from config
            options = True
        else:
            options = rabbit_ssl_options
        return options

    @pytest.mark.usefixtures("rabbit_ssl_config")
    def test_consume_over_ssl(self, container_factory, queue, rabbit_uri):

        class Service(object):
            name = "service"

            @consume(queue)
            def echo(self, payload):
                return payload

        container = container_factory(Service)
        container.start()

        publisher = PublisherCore(rabbit_uri)

        with entrypoint_waiter(container, 'echo') as result:
            publisher.publish("payload", routing_key=queue.name)
        assert result.get() == "payload"

    @pytest.mark.usefixtures("rabbit_config")
    def test_publisher_over_ssl(
        self, container_factory, queue, rabbit_ssl_uri, rabbit_ssl_options
    ):
        class PublisherService(object):
            name = "publisher"

            publish = Publisher(uri=rabbit_ssl_uri, ssl=rabbit_ssl_options)

            @dummy
            def method(self, payload):
                return self.publish(payload, routing_key=queue.name)

        class ConsumerService(object):
            name = "consumer"

            @consume(queue)
            def echo(self, payload):
                return payload

        publisher = container_factory(PublisherService)
        publisher.start()

        consumer = container_factory(ConsumerService)
        consumer.start()

        with entrypoint_waiter(consumer, 'echo') as result:
            with entrypoint_hook(publisher, 'method') as publish:
                publish("payload")
        assert result.get() == "payload"


@pytest.mark.usefixtures("rabbit_config")
def test_stop_with_active_worker(container_factory, queue_info):
    """ Test behaviour when we stop a container with an active worker.

    Expect the consumer to stop and the message be requeued, but the container
    to continue running the active worker until it completes.

    This is not desirable behaviour but it is consistent with the old
    implementation. It would be better to stop the consumer but keep the
    channel alive until the worker has completed and the message can be
    ack'd, but we can't do that with kombu or without per-entrypoint worker
    pools.
    """

    block = Event()

    class Service(object):
        name = "service"

        @consume(Queue(name="queue"))
        def method(self, payload):
            block.wait()

    container = container_factory(Service)
    container.start()

    publisher = PublisherCore(config['AMQP_URI'])
    publisher.publish("payload", routing_key="queue")

    gt = eventlet.spawn(container.stop)

    @retry
    def consumer_removed():
        info = queue_info('queue')
        assert info.consumer_count == 0
        assert info.message_count == 1

    consumer_removed()

    assert not gt.dead
    block.send(True)

    gt.wait()
    assert gt.dead


class TestEntrypointArguments:

    @pytest.mark.usefixtures("rabbit_config")
    def test_expected_exceptions_and_sensitive_arguments(
        self, container_factory
    ):

        class Boom(Exception):
            pass

        class Service(object):
            name = "service"

            @consume(
                Queue(name="queue"),
                expected_exceptions=Boom,
                sensitive_arguments=["payload"]
            )
            def method(self, payload):
                pass  # pragma: no cover

        container = container_factory(Service)
        container.start()

        entrypoint = get_extension(container, Consumer)
        assert entrypoint.expected_exceptions == Boom
        assert entrypoint.sensitive_arguments == ["payload"]



"""Provide custom code `interact` for interactive nameko shell.
"""
from __future__ import absolute_import

from code import InteractiveConsole


# This class is covered by tests run in subprocess hence the no cover
class RaisingInteractiveConsole(InteractiveConsole):  # pragma: no cover
    """ Custom InterativeConsole class that allows raising exception if needed.
    """

    def __init__(
        self, locals=None, filename="<console>", raise_expections=False
    ):
        InteractiveConsole.__init__(self, locals=locals, filename=filename)
        self.raise_expections = raise_expections

    def runcode(self, code):
        try:
            exec(code, self.locals)
        except SystemExit:
            raise
        except:
            self.showtraceback()
            if self.raise_expections:
                raise


# This function is covered by tests run in subprocess hence the no cover
def interact(
    banner=None, local=None, raise_expections=False
):  # pragma: no cover
    try:
        import readline  # noqa: F401
    except ImportError:
        pass

    console = RaisingInteractiveConsole(
        locals=local, raise_expections=raise_expections
    )
    console.interact(banner=banner)



from nameko.constants import (
    AUTH_TOKEN_CONTEXT_KEY, LANGUAGE_CONTEXT_KEY, USER_AGENT_CONTEXT_KEY,
    USER_ID_CONTEXT_KEY
)
from nameko.extensions import DependencyProvider


class ContextDataProvider(DependencyProvider):
    context_key = None

    def get_dependency(self, worker_ctx):
        return worker_ctx.data.get(self.context_key)


class Language(ContextDataProvider):
    context_key = LANGUAGE_CONTEXT_KEY


class UserId(ContextDataProvider):
    context_key = USER_ID_CONTEXT_KEY


class UserAgent(ContextDataProvider):
    context_key = USER_AGENT_CONTEXT_KEY


class AuthToken(ContextDataProvider):
    context_key = AUTH_TOKEN_CONTEXT_KEY



import pytest

from nameko.events import event_handler
from nameko.standalone.events import event_dispatcher
from nameko.testing.services import entrypoint_waiter


class ServiceB:
    """ Event listening service.
    """
    name = "service_b"

    @event_handler("service_a", "event_type")
    def handle_event(self, payload):
        print("service b received", payload)


@pytest.mark.usefixtures("rabbit_config")
def test_event_interface(container_factory):

    container = container_factory(ServiceB)
    container.start()

    dispatch = event_dispatcher()

    # prints "service b received payload" before "exited"
    with entrypoint_waiter(container, 'handle_event'):
        dispatch("service_a", "event_type", "payload")
    print("exited")



from functools import partial
from logging import getLogger

import six
from eventlet.event import Event
from werkzeug.routing import Rule
from werkzeug.wrappers import Response

from nameko.exceptions import BadRequest, serialize
from nameko.extensions import Entrypoint
from nameko.web.server import WebServer


_log = getLogger(__name__)


class HttpRequestHandler(Entrypoint):
    server = WebServer()

    def __init__(self, method, url, **kwargs):
        self.method = method
        self.url = url
        super(HttpRequestHandler, self).__init__(**kwargs)

    def get_url_rule(self):
        return Rule(self.url, methods=self.method.split(','))

    def setup(self):
        self.server.register_provider(self)

    def stop(self):
        self.server.unregister_provider(self)
        super(HttpRequestHandler, self).stop()

    def get_entrypoint_parameters(self, request):
        args = (request,)
        kwargs = request.path_values
        return args, kwargs

    def handle_request(self, request):
        request.shallow = False
        try:
            context_data = self.server.context_data_from_headers(request)
            args, kwargs = self.get_entrypoint_parameters(request)

            self.check_signature(args, kwargs)
            event = Event()
            self.container.spawn_worker(
                self, args, kwargs, context_data=context_data,
                handle_result=partial(self.handle_result, event))
            result = event.wait()

            response = self.response_from_result(result)

        except Exception as exc:
            response = self.response_from_exception(exc)
        return response

    def handle_result(self, event, worker_ctx, result, exc_info):
        event.send(result, exc_info)
        return result, exc_info

    def response_from_result(self, result):
        if isinstance(result, Response):
            return result

        headers = None
        if isinstance(result, tuple):
            if len(result) == 3:
                status, headers, payload = result
            else:
                status, payload = result
        else:
            payload = result
            status = 200

        if not isinstance(payload, six.string_types):
            raise TypeError(
                "Payload must be a string. Got `{!r}`".format(payload)
            )

        return Response(
            payload,
            status=status,
            headers=headers,
        )

    def response_from_exception(self, exc):
        if (
            isinstance(exc, self.expected_exceptions) or
            isinstance(exc, BadRequest)
        ):
            status_code = 400
        else:
            status_code = 500
        error_dict = serialize(exc)
        payload = u'Error: {exc_type}: {value}\n'.format(**error_dict)

        return Response(
            payload,
            status=status_code,
        )


http = HttpRequestHandler.decorator



from nameko.events import BROADCAST, event_handler

class ListenerService:
    name = "listener"

    @event_handler(
        "monitor", "ping", handler_type=BROADCAST, reliable_delivery=False
    )
    def ping(self, payload):
        # all running services will respond
        print("pong from {}".format(self.name))



from collections import namedtuple
from copy import deepcopy

import kombu.serialization

from nameko import config
from nameko.constants import (
    ACCEPT_CONFIG_KEY, DEFAULT_SERIALIZER, SERIALIZER_CONFIG_KEY,
    SERIALIZERS_CONFIG_KEY
)
from nameko.exceptions import ConfigurationError
from nameko.utils import import_from_path


SerializationConfig = namedtuple("SerializationConfig", 'serializer accept')


def setup():

    serializers = deepcopy(config.get(SERIALIZERS_CONFIG_KEY, {}))
    for name, kwargs in serializers.items():
        encoder = import_from_path(kwargs.pop('encoder'))
        decoder = import_from_path(kwargs.pop('decoder'))
        kombu.serialization.register(
            name, encoder=encoder, decoder=decoder, **kwargs)

    serializer = config.get(SERIALIZER_CONFIG_KEY, DEFAULT_SERIALIZER)

    accept = config.get(ACCEPT_CONFIG_KEY, [serializer])

    for name in [serializer] + accept:
        try:
            kombu.serialization.registry.name_to_type[name]
        except KeyError:
            raise ConfigurationError(
                'Please register a serializer for "{}" format'.format(name))

    return SerializationConfig(serializer, accept)



import errno
import os
import signal
import socket
from os.path import abspath, dirname, join
from textwrap import dedent

import eventlet
import pytest
from mock import call, patch

from nameko import config
from nameko.cli.run import run, setup_backdoor
from nameko.cli.utils import import_services
from nameko.constants import SERIALIZER_CONFIG_KEY, WEB_SERVER_CONFIG_KEY
from nameko.runners import ServiceRunner
from nameko.standalone.rpc import ClusterRpcClient
from nameko.testing.waiting import wait_for_call

from test.sample import Service


TEST_CONFIG_FILE = abspath(join(dirname(__file__), "config.yaml"))


@pytest.mark.usefixtures("rabbit_config")
def test_run(command):

    # start runner and wait for it to come up
    with wait_for_call(ServiceRunner, "start"):
        gt = eventlet.spawn(
            command, "nameko", "run", "--backdoor-port", "0", "test.sample:Service"
        )

    # make sure service launches ok
    with ClusterRpcClient() as client:
        client.service.ping()

    # stop service
    pid = os.getpid()
    os.kill(pid, signal.SIGTERM)
    gt.wait()


@pytest.mark.usefixtures("rabbit_config")
def test_main_with_config(command, tmpdir):

    config_file = tmpdir.join("config.yaml")
    config_file.write(
        """
        WEB_SERVER_ADDRESS: '0.0.0.0:8001'
        serializer: 'json'
    """
    )

    assert WEB_SERVER_CONFIG_KEY not in config
    assert SERIALIZER_CONFIG_KEY not in config

    with patch("nameko.cli.run.main") as run:

        command("nameko", "run", "--config", config_file.strpath, "test.sample")

        assert run.call_count == 1

    assert config[WEB_SERVER_CONFIG_KEY] == "0.0.0.0:8001"
    assert config[SERIALIZER_CONFIG_KEY] == "json"


@pytest.mark.usefixtures("rabbit_config")
def test_main_with_config_options(command, tmpdir):

    config_file = tmpdir.join("config.yaml")
    config_file.write(
        """
        WEB_SERVER_ADDRESS: '0.0.0.0:8001'
        serializer: 'json'
    """
    )

    assert WEB_SERVER_CONFIG_KEY not in config
    assert SERIALIZER_CONFIG_KEY not in config
    assert "EGG" not in config

    with patch("nameko.cli.run.main") as run:

        command(
            "nameko",
            "run",
            "--config",
            config_file.strpath,
            "--define",
            "serializer=pickle",
            "--define",
            'EGG=[{"spam": True}]',
            "test.sample",
        )

        assert run.call_count == 1

    assert config[WEB_SERVER_CONFIG_KEY] == "0.0.0.0:8001"
    assert config[SERIALIZER_CONFIG_KEY] == "pickle"
    assert config["EGG"] == [{"spam": True}]


@pytest.mark.usefixtures("rabbit_config")
def test_main_with_logging_config(command, tmpdir):

    config_content = """
        AMQP_URI: {amqp_uri}
        LOGGING:
            version: 1
            disable_existing_loggers: false
            formatters:
                simple:
                    format: "%(name)s - %(levelname)s - %(message)s"
            handlers:
                capture:
                    class: logging.FileHandler
                    level: INFO
                    formatter: simple
                    filename: {capture_file}
            root:
                level: INFO
                handlers: [capture]
    """

    capture_file = tmpdir.join("capture.log")

    config_file = tmpdir.join("config.yaml")
    config_file.write(
        dedent(
            config_content.format(
                capture_file=capture_file.strpath, amqp_uri=config["AMQP_URI"]
            )
        )
    )

    # start runner and wait for it to come up
    with wait_for_call(ServiceRunner, "start"):
        gt = eventlet.spawn(
            command, "nameko", "run", "--config", config_file.strpath, "test.sample"
        )

    with ClusterRpcClient() as client:
        client.service.ping()

    pid = os.getpid()
    os.kill(pid, signal.SIGTERM)
    gt.wait()

    assert "test.sample - INFO - ping!" in capture_file.read()


def test_multiple_service_modules(command):

    with patch("nameko.cli.run.main") as run:

        command(
            "nameko",
            "run",
            "test.foo",
            "test.bar"
        )

        assert run.call_args_list == [call(("test.foo", "test.bar"), None)]


def test_parse_config_before_service_import(command, tmpdir, add_to_sys_path):

    config_file = tmpdir.join("config.yaml")
    config_file.write(dedent(
        """
        NAME: dynamic
        """
    ))

    service_file = tmpdir.join("service.py")
    service_file.write(dedent(
        """
        from nameko import config
        from nameko.testing.services import dummy

        class Service:
            name = config.get("NAME", "notfound")

            @dummy
            def method(self):
                return "OK"

        """
    ))

    service_names = []

    def cb(args, kwargs, res, exc_info):
        service_runner, = args
        service_names.extend(service_runner.service_names)
        return True

    with add_to_sys_path(tmpdir.strpath):

        with wait_for_call(ServiceRunner, "start", callback=cb):
            eventlet.spawn(
                command, "nameko", "run", "--config", config_file.strpath,
                "service:Service"
            )

    assert service_names == ["dynamic"]


def test_import_ok():
    assert import_services("test.sample") == [Service]
    assert import_services("test.sample:Service") == [Service]


def test_import_missing():
    with pytest.raises(ValueError) as exc:
        import_services("non_existent")
    assert "No module named" in str(exc.value)
    assert "non_existent" in str(exc.value)


def test_import_filename():
    with pytest.raises(ValueError) as exc:
        import_services("test/sample.py")
    assert "did you mean 'test.sample'?" in str(exc)


def test_import_broken():
    with pytest.raises(ValueError):
        import_services("test.broken_sample")


def test_import_missing_class():
    with pytest.raises(ValueError) as exc:
        import_services("test.sample:NonExistent")
    assert "Failed to find service class" in str(exc)


def test_import_not_a_class():
    with pytest.raises(ValueError) as exc:
        import_services("test.sample:rpc")
    assert "Service must be a class" in str(exc)


def test_import_no_service_classes():
    with pytest.raises(ValueError):
        import_services("test")


def recv_until_prompt(sock):
    data = b""
    part = b""
    while not data[-5:] == b"\n>>> ":
        part = sock.recv(4096)
        data += part
    return data


def test_backdoor():
    runner = object()
    green_socket, gt = setup_backdoor(runner, ("localhost", 0))
    eventlet.sleep(0)  # give backdoor a chance to spawn
    socket_name = green_socket.fd.getsockname()
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(socket_name)
    recv_until_prompt(sock)  # banner

    sock.sendall(b"runner\n")
    runner_repr = recv_until_prompt(sock)
    assert repr(runner) in str(runner_repr)

    sock.sendall(b"quit()\n")
    error = recv_until_prompt(sock)
    assert "RuntimeError: This would kill your service" in str(error)
    sock.close()
    gt.kill()


def test_stopping(rabbit_config):
    with patch("nameko.cli.run.eventlet") as mock_eventlet:
        # this is the service "runlet"
        mock_eventlet.spawn().wait.side_effect = [
            KeyboardInterrupt,
            None,  # second wait, after stop() which returns normally
        ]
        gt = eventlet.spawn(run, ["test.sample"])
        gt.wait()
        # should complete


def test_stopping_twice(rabbit_config):
    with patch("nameko.cli.run.eventlet") as mock_eventlet:
        # this is the service "runlet"
        mock_eventlet.spawn().wait.side_effect = [
            KeyboardInterrupt,
            None,  # second wait, after stop() which returns normally
        ]
        with patch("nameko.cli.run.ServiceRunner") as runner_cls:
            runner = runner_cls()
            runner.stop.side_effect = KeyboardInterrupt
            runner.kill.return_value = None

            gt = eventlet.spawn(run, ["test.sample"])
            gt.wait()


def test_os_error_for_signal(rabbit_config):
    with patch("nameko.cli.run.eventlet") as mock_eventlet:
        # this is the service "runlet"
        mock_eventlet.spawn().wait.side_effect = [
            OSError(errno.EINTR, ""),
            None,  # second wait, after stop() which returns normally
        ]
        # don't actually start the service -- we're not firing a real signal
        # so the signal handler won't stop it again
        with patch.object(ServiceRunner, "start"):
            gt = eventlet.spawn(run, ["test.sample"])
            gt.wait()
        # should complete


def test_other_errors_propagate(rabbit_config):
    with patch("nameko.cli.run.eventlet") as mock_eventlet:
        # this is the service "runlet"
        mock_eventlet.spawn().wait.side_effect = [
            OSError(0, ""),
            None,  # second wait, after stop() which returns normally
        ]
        # don't actually start the service -- there's no real OSError that
        # would otherwise kill the whole process
        with patch.object(ServiceRunner, "start"):
            gt = eventlet.spawn(run, ["test.sample"])
            with pytest.raises(OSError):
                gt.wait()


@pytest.mark.usefixtures("empty_config")
def test_broker_option_deprecated(command, rabbit_uri):

    with patch("nameko.cli.run.run") as run:
        with patch("nameko.cli.utils.config.warnings") as warnings:
            command("nameko", "run", "--broker", rabbit_uri, "test.sample")

    assert run.call_count == 1

    assert warnings.warn.call_count
